import os

from PyQt5 import QtCore, QtWebEngineWidgets, QtWidgets

no_diff_html = (
    " <!DOCTYPE html><html><body><p>No differences to display between"
    " Source A and Source B.</p></body></html> "
)
no_file_html = (
    " <!DOCTYPE html><html><body><p>Please select a diff to display from"
    " the list.</p></body></html> "
)


class Ui_ViewerDialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Comparison Viewer")
        Dialog.resize(1800, 800)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setGeometry(QtCore.QRect(1450, 762, 341, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Close)
        self.buttonBox.setObjectName("buttonBox")
        self.listWidget = QtWidgets.QListWidget(Dialog)
        self.listWidget.setGeometry(QtCore.QRect(10, 10, 270, 500))
        self.listWidget.setObjectName("listWidget")
        for i in range(0, self.numFiles):
            item = QtWidgets.QListWidgetItem()
            self.listWidget.addItem(item)
        self.webEngineView = QtWebEngineWidgets.QWebEngineView(Dialog)
        self.webEngineView.setHtml(no_file_html)
        self.webEngineView.setGeometry(QtCore.QRect(290, 10, 1500, 740))
        self.webEngineView.setObjectName("widget")

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        self.listWidget.itemPressed["QListWidgetItem*"].connect(self.openFile)
        QtCore.QMetaObject.connectSlotsByName(Dialog)
        if self.path is None:
            self.path = "/does/not/exist"

    def setFileList(self, filelist):
        self.filelist = filelist
        self.numFiles = len(filelist)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Comparison Viewer", "Comparison Viewer"))
        __sortingEnabled = self.listWidget.isSortingEnabled()
        self.listWidget.setSortingEnabled(False)

        for i in range(0, self.numFiles):
            item = self.listWidget.item(i)
            # listsplt = self.filelist[i].split("/")
            name = self.filelist[i].replace(self.path, "")
            rmExtName = name.replace(".html", "")
            item.setText(_translate("Dialog", rmExtName))
        self.listWidget.setSortingEnabled(__sortingEnabled)

    def openFile(self):
        for i in range(0, self.numFiles):
            if self.listWidget.currentRow() == i:
                self.updateWebView(self.filelist[i])
                break

    def updateWebView(self, newpath):
        if os.path.exists(newpath):
            self.webEngineView.load(QtCore.QUrl().fromLocalFile(newpath))
        else:
            self.webEngineView.setHtml(no_diff_html)

    def setPath(self, newpath):
        self.path = newpath


class ComparisonViewer(Ui_ViewerDialog):
    def __init__(self, parent):
        self.dialog = QtWidgets.QDialog()

    def setup(self, rootPath, fileList):
        self.ui = Ui_ViewerDialog()
        self.ui.setPath(rootPath)
        self.ui.setFileList(fileList)
        self.ui.setupUi(self.dialog)

    def show(self):
        self.dialog.show()


if __name__ == "__main__":
    import sys

    app = QtWidgets.QApplication(sys.argv)
    Dialog = QtWidgets.QDialog()
    ui = Ui_ViewerDialog()
    ui.setPath("rootPath")
    ui.setFileList([])
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
from PyQt5.QtWidgets import QDialog

from dls_powerpmacanalyse.ui_formLogin import Ui_Login


class Loginform(QDialog, Ui_Login):
    def __init__(self, parent):
        QDialog.__init__(self, parent)
        self.setupUi(self)
        self.parent = parent
        self.username = None
        self.password = None

    def clickedOK(self):
        self.username = self.lneUsername.text()
        self.password = self.lnePassword.text()
        self.lneUsername.setText("")
        self.lnePassword.setText("")
        self.accept()

    def clickedCancel(self):
        self.lneUsername.setText("")
        self.lnePassword.setText("")
        self.reject()



# Compute a version number from a git repo or archive

# This file is released into the public domain. Generated by:
# versiongit-0.6 (https://github.com/dls-controls/versiongit)
import os
import re
import sys
from subprocess import STDOUT, CalledProcessError, check_output

# These will be filled in if git archive is run or by setup.py cmdclasses
GIT_REFS = "$Format:%D$"
GIT_SHA1 = "$Format:%h$"


def get_version_from_git(path=None):
    """Try to parse version from git describe, fallback to git archive tags"""
    tag, plus, suffix = "0", "untagged", ""
    if not GIT_SHA1.startswith("$"):
        # git archive or the cmdclasses below have filled in these strings
        sha1 = GIT_SHA1
        for ref_name in GIT_REFS.split(", "):
            if ref_name.startswith("tag: "):
                # git from 1.8.3 onwards labels archive tags "tag: TAGNAME"
                tag, plus = ref_name[5:], "0"
    else:
        if path is None:
            # If no path to git repo, choose the directory this file is in
            path = os.path.dirname(os.path.abspath(__file__))
        git_cmd = "git describe --tags --dirty --always --long".split()
        # output is TAG-NUM-gHEX[-dirty] or HEX[-dirty]
        try:
            out = check_output(git_cmd, stderr=STDOUT, cwd=path).decode().strip()
        except Exception as e:
            sys.stderr.write("%s: %s\n" % (type(e).__name__, str(e)))
            if isinstance(e, CalledProcessError):
                sys.stderr.write("-> %s" % e.output.decode())
            return "0+unknown", None, e
        else:
            if out.endswith("-dirty"):
                out = out[:-6]
                suffix = ".dirty"
            if "-" in out:
                # There is a tag, extract it and the other pieces
                match = re.search(r"^(.+)-(\d+)-g([0-9a-f]+)$", out)
                tag, plus, sha1 = match.groups()
            else:
                # No tag, just sha1
                sha1 = out
    # Replace dashes in tag for dots
    # Remove this line when we stop supporting python 2.7
    tag = tag.replace("-", ".")
    if plus != "0" or suffix:
        # Not on a tag, add additional info
        tag = "%(tag)s+%(plus)s.g%(sha1)s%(suffix)s" % locals()
    return tag, sha1, None


__version__, git_sha1, git_error = get_version_from_git()


def get_cmdclass(build_py=None, sdist=None):
    """Create cmdclass dict to pass to setuptools.setup that will write a
    _version_static.py file in our resultant sdist, wheel or egg"""
    if build_py is None:
        from setuptools.command.build_py import build_py
    if sdist is None:
        from setuptools.command.sdist import sdist

    def make_version_static(base_dir, pkg):
        vg = os.path.join(base_dir, pkg.split(".")[0], "_version_git.py")
        if os.path.isfile(vg):
            lines = open(vg).readlines()
            with open(vg, "w") as f:
                for line in lines:
                    # Replace GIT_* with static versions
                    if line.startswith("GIT_SHA1 = "):
                        f.write("GIT_SHA1 = '%s'\n" % git_sha1)
                    elif line.startswith("GIT_REFS = "):
                        f.write("GIT_REFS = 'tag: %s'\n" % __version__)
                    else:
                        f.write(line)

    class BuildPy(build_py):
        def run(self):
            build_py.run(self)
            for pkg in self.packages:
                make_version_static(self.build_lib, pkg)

    class Sdist(sdist):
        def make_release_tree(self, base_dir, files):
            sdist.make_release_tree(self, base_dir, files)
            for pkg in self.distribution.packages:
                make_version_static(base_dir, pkg)

    return dict(build_py=BuildPy, sdist=Sdist)



import argparse
import difflib
import logging
import os
import re
import sys
import time
from difflib import HtmlDiff

import numpy as np
from dls_pmaclib import dls_pmacremote
from scp import SCPClient

import dls_powerpmacanalyse.ppmacanalyse_control


def timer(func):
    def measureExecutionTime(*args, **kwargs):
        startTime = time.time()
        result = func(*args, **kwargs)
        logging.info(
            "Processing time of %s(): %.2f seconds."
            % (func.__qualname__, time.time() - startTime)
        )
        return result

    return measureExecutionTime


def connectDisconnect(func):
    def wrapped_func(self, *args):
        sshClient.port = self.port
        sshClient.hostname = self.ipAddress
        sshClient.connect()
        func(self, *args)
        sshClient.disconnect()

    return wrapped_func


def fileExists(file):
    return os.path.isfile(file)


def parseArgs():
    parser = argparse.ArgumentParser(
        description="Interact with a Power PMAC",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "-i",
        "--interface",
        metavar="",
        nargs=1,
        help=(
            "Network interface address of Power PMAC.\n--interface <ip address>:<port>."
        ),
    )
    parser.add_argument(
        "-b",
        "--backup",
        metavar="",
        nargs="*",
        help=(
            "Back-up configuration of Power PMAC.\n"
            "--backup <type>/<type> <ignore>\n"
            "<type> = all/active/project\n"
            "<ignore> = path to ignore file"
        ),
    )
    parser.add_argument(
        "-r",
        "--recover",
        metavar="",
        nargs=1,
        help=(
            "Recover configuration of Power PMAC using the recovery"
            "stick method.\n"
            "--recover <usrflash dir>\n"
            "<usrflash dir> = local copy of the /opt/ppmac/usrflash\n"
            "directory on a Power PMAC."
        ),
    )
    parser.add_argument(
        "-c",
        "--compare",
        metavar="",
        nargs="*",
        help=(
            "Compare configuration of two Power PMACs.\n--compare <source_a> <source_b>"
            " <ignore_file>\n<source_a> and <source_b> define"
            " the two sources to compare. They can \ntake the form of a path to a"
            " back-up directory or a network interface <ip_address>:<port>.\n<ignore"
            " file> is the path to the file listing which data structures should be"
            " ignored."
        ),
    )
    parser.add_argument(
        "-d",
        "--download",
        metavar="",
        nargs=1,
        help=(
            "Download configuration onto Power PMAC.\n"
            "--download <usrflash dir>\n"
            "<project dir> = local copy of the /var/ftp/usrflash/Project\n"
            "directory on a Power PMAC."
        ),
    )
    parser.add_argument(
        "-f",
        "--resultsdir",
        metavar="",
        nargs=1,
        help=(
            "Directory in which to place output of analysis.\n"
            "--resultsdir <results dir>"
        ),
    )
    parser.add_argument(
        "-g",
        "--gui",
        action="store_true",
        help=("Launch Power PMAC analyse GUI."),
    )
    parser.add_argument(
        "-u",
        "--username",
        metavar="",
        nargs=1,
        help=("Power PMAC username"),
    )
    parser.add_argument(
        "-p",
        "--password",
        metavar="",
        nargs=1,
        help=("Power Pmac password"),
    )
    parser.add_argument("-n", "--name", metavar="", nargs=1, help="Name of Power PMAC.")
    return parser.parse_args()


def isValidNetworkInterface(interface):
    interfaceSplit = interface.split(":")
    if len(interfaceSplit) != 2:
        return False
    ipAddress = interfaceSplit[0]
    port = interfaceSplit[1]
    isValid = False
    if "." in ipAddress:
        elems = ipAddress.strip().split(".")
        if len(elems) == 4:
            for i in elems:
                if i.isnumeric() and int(i) >= 0 and int(i) <= 255:
                    isValid = True
                else:
                    isValid = False
                    break
    if not port.isnumeric():
        isValid = False
    return isValid


def exitGpascii():
    (exitGpascii, status) = sshClient.sendCommand("\x03")
    if not status:
        raise IOError("Failed to exit gpascii.")


def executeRemoteShellCommand(cmd):
    """
    Execute a command on a remote server. The call to stdout.channel.recv_exit_status()
    blocks until the resulting
    process on the remote server has finished.
    :param cmd: string containing command to be executed.
    :return:
    """
    logging.info(f"Executing '{cmd}' on remote server...")
    stdin, stdout, stderr = sshClient.client.exec_command(cmd)
    if stdout.channel.recv_exit_status() != 0:
        raise RuntimeError(f"Error executing command {cmd} on remote machine.")
    logging.info(f"Command wrote to stdout '{stdout.read()}'.")


def scpFromPowerPMACtoLocal(source, destination, recursive=True):
    logging.info(
        f"scp files/dirs '{source}' from remote server into local dir '{destination}'."
    )
    try:
        scp = SCPClient(sshClient.client.get_transport(), sanitize=lambda x: x)
        scp.get(source, destination, recursive)
        scp.close()
    except Exception as e:
        print(f"Error: {e}, unable to get directory from remote host: {source}")


def scpFromLocalToPowerPMAC(files, remote_path, recursive=False):
    logging.info(
        f"scp files/dirs '{files}' into remote server directory '{remote_path}'."
    )
    try:
        scp = SCPClient(sshClient.client.get_transport(), sanitize=lambda x: x)
        scp.put(files, remote_path, recursive)
        scp.close()
    except Exception as e:
        print(f"Error: {e}, unable to copy files {files} to remote path: {remote_path}")


def nthRepl(s, sub, repl, nth):
    find = s.find(sub)
    i = find != -1
    while find != -1 and i != nth:
        find = s.find(sub, find + 1)
        i += 1
    if i == nth:
        return s[:find] + repl + s[find + len(sub) :]
    return s


def find_nth(s, sub, n):
    start = s.find(sub)
    while start >= 0 and n > 1:
        start = s.find(sub, start + len(sub))
        n -= 1
    return start


def responseListToDict(responseList, splitChars="="):
    responseDict = {}
    if responseList != ["", ""]:
        for element in responseList:
            nameVal = element.split(splitChars)
            responseDict[nameVal[0]] = nameVal[1]
    return responseDict


def mergeDicts(*dict_args):
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def comparedicts(d1, d2):
    d1Keys = set(d1.keys())
    d2Keys = set(d2.keys())
    shared_keys = d1Keys.intersection(d2Keys)
    added = d1Keys - d2Keys
    removed = d2Keys - d1Keys
    modified = {o: (d1[o], d2[o]) for o in shared_keys if d1[o] != d2[o]}
    same = set(o for o in shared_keys if d1[o] == d2[o])
    return added, removed, modified, same


recoveryCmds = """
#!/bin/bash

echo "Entering /tmp/recover." > /tmp/recover.log 2>&1
cd /tmp/recover
if [[ $? -ne 0 ]] ; then
    echo "Unable to cd into /tmp/recover. Exiting." >> /tmp/recover.log 2>&1
    exit 1
fi

echo "Starting sync." >> /tmp/recover.log 2>&1
echo "Mounting /opt file system as read/write" >> /tmp/recover.log 2>&1
mount -o remount,rw /opt/ >> /tmp/recover.log 2>&1

# Copy backup files into /opt/ppmac/usrflash
echo "Copying backup files into /opt/ppmac/usrflash/" >> /tmp/recover.log
cp -a * /opt/ppmac/usrflash/ >> /tmp/recover.log 2>&1

echo "Syncing file system." >> /tmp/recover.log 2>&1
sync >> /tmp/recover.log 2>&1

echo "Mounting /opt file system as read only." >> /tmp/recover.log 2>&1
mount -o remount,ro /opt/ >> /tmp/recover.log 2>&1

echo "Sync completed." >> /tmp/recover.log 2>&1
"""


class PPMACLexer(object):
    # Tokens formed from the 'Power PMAC on-line commands' and
    # 'Power PMAC program commands'.
    # As given in the Power PMAC Software reference manual 22nd Mar 2021 p.40-47
    commandTokens = {
        "a",
        "aa",
        "abort",
        "abs",
        "adisable",
        "alias",
        "all",
        "apc",
        "assign",
        "b",
        "bb",
        "begin",
        "bgcplc",
        "break",
        "bstart",
        "bstop",
        "buffers",
        "c",
        "cc",
        "call",
        "callsub",
        "case",
        "ccall",
        "cclr",
        "ccmode0",
        "ccmode1",
        "ccmode2",
        "ccmode3",
        "ccr",
        "cdef",
        "cexec",
        "cipserialnumber",
        "cipvendorcode",
        "circle",
        "clear",
        "close",
        "clrf",
        "cmd",
        "config",
        "continue",
        "cout",
        "cpu",
        "cpx",
        "cset",
        "cskip",
        "cundef",
        "cx",
        "d",
        "dd",
        "date",
        "ddisable",
        "define",
        "delete",
        "disable",
        "dkill",
        "do",
        "dread",
        "dtogread",
        "dwell",
        "ecat",
        "echo",
        "else",
        "enable",
        "f",
        "ff",
        "fload",
        "forward",
        "frax",
        "frax2",
        "fread",
        "free",
        "fsave",
        "g",
        "gg",
        "gather",
        "gosub",
        "goto",
        "h",
        "hh",
        "halt",
        "hm",
        "hmz",
        "hold",
        "home",
        "homez",
        "i",
        "if",
        "inc",
        "inverse",
        "j",
        "jog",
        "jogret",
        "k",
        "kill",
        "l",
        "ll",
        "lh",
        "lhpurge",
        "linear",
        "list",
        "lookahead",
        "lotnum",
        "m",
        "mm",
        "macroauxiliary",
        "macroauxiliaryread",
        "macroauxiliarywrite",
        "macrocontrolledetect",
        "macrocontrollerinit",
        "macromaster",
        "macromasterread",
        "macromasterwrite",
        "macroport",
        "macroportclose",
        "macroportstate",
        "macroring",
        "macroringmasterslave",
        "macroringmasterslaveread",
        "macroringmasterslavewrite",
        "macroringorder",
        "macroringorderbackup",
        "macroringorderclrf",
        "macroringorderdetect",
        "macroringorderinit",
        "macroringorderload",
        "macroringorderrepair",
        "macroringorderrestore",
        "macroringordersave",
        "macroringorderstations",
        "macroringordersync",
        "macroringorderverify",
        "macroringorderverifybackup",
        "macroslave",
        "macroslaveread",
        "macroslavewrite",
        "macrostation",
        "macrostationclearerrors",
        "macrostationclose",
        "macrostationenable",
        "macrostationerrors",
        "macrostationfrequency",
        "macrostationringcheck",
        "macrostationstatus",
        "macrostationtype",
        "n",
        "nn",
        "nofrax",
        "nofrax2",
        "nop",
        "normal",
        "nxyz",
        "oo",
        "open",
        "out",
        "p",
        "pp",
        "pause",
        "pc",
        "pclear",
        "phase",
        "plc",
        "pload",
        "pmatch",
        "pread",
        "productcode",
        "prog",
        "pset",
        "pstore",
        "pvt",
        "q",
        "qq",
        "r",
        "rr",
        "rapid",
        "read",
        "reboot",
        "reset",
        "resetverbose",
        "resume",
        "return",
        "rotary",
        "rotfree",
        "rotfreeall",
        "rtiplc",
        "run",
        "s",
        "ss",
        "save",
        "send",
        "sendall",
        "sendallcmds",
        "sendallsystemcmds",
        "serialnum",
        "setverbose",
        "size",
        "slaves",
        "spline",
        "start",
        "step",
        "step",
        "stop",
        "string",
        "subprog",
        "switch",
        "system",
        "t",
        "tt",
        "ta",
        "td",
        "tm",
        "tread",
        "ts",
        "tsd",
        "tsel",
        "txyz",
        "txyzscale",
        "type",
        "uu",
        "undefine",
        "v",
        "vv",
        "vers",
        "vread",
        "ww",
        "xx",
        "yy",
        "z",
        "#",
        "->",
        "$",
        "$$",
        "$$$",
        "$$$***",
        "%",
        "&",
        "\\",
        "?",
        ":",
    }
    arithmeticOperatorTokens = {"+", "-", "*", "/", "%"}
    bitByBitLogicalOperatorTokens = {"&", "|", "^", "~", "<<", ">>"}
    standardAssignmentOperatorTokens = {
        "=",
        "+=",
        "-=",
        "*=",
        "/=",
        "%=",
        "&=",
        "|=",
        "^=",
        ">>=",
        "<<=",
        "++",
        "--",
    }
    synchronousAssignmentOperatorTokens = {
        "==",
        "+==",
        "-==",
        "*==",
        "/==",
        "&==",
        "|==",
        "^==",
        "++=",
        "--=",
    }
    conditionalComparatorTokens = {
        "==",
        "!=",
        "<",
        ">",
        "<=",
        ">=",
        "~",
        "!~",
        "<>",
        "!>",
        "!<",
    }
    conditionalCombinatorialOperatorTokens = {"&&", "||", "!"}
    otherTokens = {"(", ")", "{", "}", "..", ","}
    # Full list of PPMAC tokens
    tokens = (
        commandTokens.union(arithmeticOperatorTokens)
        .union(bitByBitLogicalOperatorTokens)
        .union(standardAssignmentOperatorTokens)
        .union(synchronousAssignmentOperatorTokens)
        .union(conditionalComparatorTokens)
        .union(conditionalCombinatorialOperatorTokens)
        .union(otherTokens)
    )
    spaces = " \n\t\r"
    nonAlphaNumeric = {
        "=",
        "+",
        "-",
        "*",
        "/",
        "%",
        "<",
        ">",
        "&",
        "|",
        "^",
        "!",
        "~",
        "$",
        "#",
        "?",
        ":",
        "\\",
        "(",
        ")",
        "}",
        "{",
        ".",
        ",",
    }

    class Chars(object):
        def __init__(self, chars):
            self.chars = chars
            self._chars = ""

        def isEmpty(self):
            if len(self.chars) == 0:
                return True
            else:
                return False

        def peekNext(self):
            return self.chars[1]

        def peek(self):
            return self.chars[0]

        def moveNext(self):
            c = self.chars[0]
            self._chars += self.chars[0]
            self.chars = self.chars[1:]
            return c

        def rewind(self, pos):
            if pos > 0:
                self.chars = self._chars[-pos:] + self.chars
                self._chars = self._chars[0:-pos]

    def __init__(self, chars, extendedTokens={""}):
        self.extendedTokens = {token.lower() for token in extendedTokens}
        self.tokens = []
        self.chars = self.Chars(chars.lower())
        for token in self.lex(self.chars):
            self.tokens.append(token)

    def pop(self, n=0):
        token = self.tokens[n]
        self.tokens = self.tokens[n + 1 :]
        return token

    def getTokenTypes(self):
        return [token[0] for token in self.tokens]

    def getTokenValues(self):
        return [token[1] for token in self.tokens]

    def getTokensAsString(self):
        return "".join([token[1] for token in self.tokens])

    def lex(self, chars):
        while not chars.isEmpty():
            c = chars.moveNext()
            if c in PPMACLexer.spaces:
                pass
            elif c == "$":
                yield ("hex", self.scanHexadecimal(c, chars))
            elif c in ("'", '"'):
                yield ("string", self.scanString(c, chars))
            elif re.match("[.0-9]", c):
                yield ("number", self.scanNumber(c, chars))
            elif re.match("[a-zA-Z]", c):
                yield ("symbol", self.scanSymbol(c, chars))
            elif c in PPMACLexer.nonAlphaNumeric:
                yield ("", self.scanNonAlphaNumeric(c, chars))
            else:
                raise IOError(f"Unknown token type {c}")

    def scanHexadecimal(self, c, chars):
        ret = c
        if chars.isEmpty():
            return ret
        next = chars.peek()
        allowed = "[a-f0-9]"
        while re.match(allowed, next) is not None:
            ret += chars.moveNext()
            if chars.isEmpty():
                break
            next = chars.peek()
        return ret

    def scanNonAlphaNumeric(self, c, chars):
        ret = c
        if chars.isEmpty():
            return ret
        existingToken = ""
        if ret in PPMACLexer.tokens:
            existingToken = ret
        next = chars.peek()
        allowed = PPMACLexer.nonAlphaNumeric
        while next in allowed:
            ret += chars.moveNext()
            if ret in PPMACLexer.tokens:
                existingToken = ret
            if chars.isEmpty():
                break
        if len(existingToken) > 0:
            chars.rewind(len(ret) - len(existingToken))
            ret = existingToken
        return ret

    def scanSymbol(self, c, chars):
        ret = c
        if chars.isEmpty():
            return ret
        existingToken = ""
        if ret in PPMACLexer.tokens:
            existingToken = ret
        next = chars.peek()
        # if c in "PLQRCMDI" and next.isdigit():
        #    # We have a P/L/Q/M/R/C variable
        #    allowed = "[0-9]"
        #    while re.match(allowed, next) != None:
        #        ret += chars.moveNext()
        #        if chars.isEmpty():
        #            break
        #        next = chars.peek()
        # else:
        if True:
            # Check to see if we can find an existing token in the subsequent
            # characters. If we do, take the longest existing token we find.
            # If we don't, assume the token can take the form of an active element name.
            allowed_ = "[a-zA-Z.\\[]"
            # allowed_ = "[a-zA-Z0-9.\[]"
            allowed = allowed_
            while re.match(allowed, next) is not None:
                if next == "[":
                    allowed = "[0-9\\]]"
                if next == "]":
                    allowed = allowed_
                ret += chars.moveNext()
                if ret in PPMACLexer.tokens:
                    existingToken = ret
                # Use this condition to catch active element names
                if re.sub("\\[([0-9]+)\\]", "[]", ret) in self.extendedTokens:
                    existingToken = ret
                if chars.isEmpty():
                    break
                # Use this condition to catch the few PPMACLexer.tokens that end
                # in a digit, e.g. frax2
                if (
                    re.match("[0-9]", chars.peek())
                    and ret + chars.peek() in PPMACLexer.tokens
                ):
                    existingToken = ret + chars.moveNext()
                    break
                next = chars.peek()
            if len(existingToken) > 0:
                chars.rewind(len(ret) - len(existingToken))
                ret = existingToken
        return ret

    def scanNumber(self, c, chars):
        ret = c
        if chars.isEmpty():
            return ret
        next = chars.peek()
        allowed = "[.0-9]"
        while re.match(allowed, next) is not None:
            # to catch '..' used to indicate ranges
            if next == ".":
                if chars.peekNext() == ".":
                    break
            ret += chars.moveNext()
            if chars.isEmpty():
                break
            next = chars.peek()
        return ret

    def scanMathsChars(self, c, chars):
        ret = c
        if chars.isEmpty():
            return ret
        while ret + chars.peek() in PPMACLexer.maths:
            c = chars.moveNext()
            ret += c
            if chars.isEmpty():
                break
        return (self.mathsDict[ret], ret)

    def scanString(self, delim, chars):
        ret = ""
        next = chars.peek()
        while next != delim:
            ret += chars.moveNext()
            if chars.isEmpty():
                break
            next = chars.peek()
        chars.moveNext()
        return ret


class PPMACProject(object):
    """
    Class containing files and directories included in a project
    """

    class Directory(object):
        def __init__(self, path, files):
            self.path = path
            # self.subdirs = {}  # dict of directory objects
            self.files = files  # dict of files

    class File(object):
        def __init__(self, name, dir, proj):
            self.name = name
            self.dir = dir  # directory object
            filePath = f"{dir}/{name}"
            self.extension = os.path.splitext(filePath)[1]
            self.contents = proj.getFileContents(filePath)

    def __init__(self, source, root, tempDir=None):
        # absolute path the project directory
        self.root = root
        # Source of project (hardware or repo)
        self.source = source
        # Dictionary of files contained in the project
        self.files = {}
        # Dictionary of directories contained in the project
        self.dirs = {}
        if self.source == "hardware":
            if root.find("usrflash") == -1:
                raise RuntimeError(
                    f'Root directory "{root}" invalid: not a project directory.'
                )
            if tempDir is None:
                raise RuntimeError(
                    "Please define a temporary directory into which the project from "
                    "the ppmac will be copied."
                )
            self.root = tempDir
            os.makedirs(tempDir, exist_ok=True)
            scpFromPowerPMACtoLocal(source=root, destination=tempDir, recursive=True)
        elif self.source != "repository":
            raise RuntimeError(
                'Invalid project source: should be "hardware" or "repository".'
            )
        self.buildProjectTree(self.root)

    def buildProjectTree(self, start):
        for root, dirs, files in os.walk(start):
            root_ = root.replace(start, "", 1)
            for name in dirs:
                dirName = os.path.join(root_, name)
                self.dirs[dirName] = self.Directory(dirName, files)
            for name in files:
                fileName = os.path.join(root_, name)
                self.files[fileName] = self.File(name, root_, self)

    def getFileContents(self, file):
        contents = []
        file = f"{self.root}/{file}"
        with open(file, "r", encoding="ISO-8859-1") as readFile:
            for line in readFile:
                contents.append(line)
        return contents


class ProjectCompare(object):
    """
    Compare two project filesystems
    """

    class FileDiffs(object):
        """
        Object holding the differences between two project files
        """

        def __init__(self, fileA, fileB):
            self.fileA = fileA
            self.fileB = fileB
            self.same = fileA.sha256 == fileB.sha256
            # plus some object to hold the line-by-line differences

    def __init__(self, projectA, projectB, ignore=None):
        self.projectA = projectA
        self.projectB = projectB
        self.filesOnlyInA = {}
        self.filesOnlyInB = {}
        self.filesInAandB = {}

    def setProjectA(self, project):
        self.projectA = project

    def setprojectB(self, project):
        self.projectB = project

    def compareProjectFiles(self, diffDirPath):
        fileNamesA = set(self.projectA.files.keys())
        fileNamesB = set(self.projectB.files.keys())
        fileNamesOnlyInA = fileNamesA - fileNamesB
        fileNamesOnlyInB = fileNamesB - fileNamesA
        fileNamesInAandB = fileNamesA & fileNamesB
        self.filesOnlyInA = self.filesOnlyInB = {}
        self.filesOnlyInA = {
            fileName: self.projectA.files[fileName] for fileName in fileNamesOnlyInA
        }
        self.filesOnlyInB = {
            fileName: self.projectB.files[fileName] for fileName in fileNamesOnlyInB
        }
        for fileName in fileNamesInAandB:
            self.filesInAandB[fileName] = {
                "A": self.projectA.files[fileName],
                "B": self.projectB.files[fileName],
            }
        if len(self.filesOnlyInA) or len(self.filesOnlyInB):
            os.makedirs(diffDirPath, exist_ok=True)
            compArr1 = []
            compArr2 = []
            for name in fileNamesA:
                compArr1.append(name)
            for name in fileNamesB:
                compArr2.append(name)
            compArr1.sort()
            compArr2.sort()

            htmlDiff = HtmlDiff()
            difference = htmlDiff.make_file(
                compArr1,
                compArr2,
                fromdesc="Programs in Source A",
                todesc="Programs in Source B",
                context=False,
            )
            diffFile = f"{diffDirPath}/MissingFiles.html"
            with open(diffFile, "w") as f:
                for line in difference.splitlines():
                    print(line, file=f)

            missingFile = f"{diffDirPath}/missingFiles.txt"
            with open(missingFile, "w+") as writeFile:
                writeFile.write(
                    f"@@ Project files in source '{self.projectA.source}' but"
                    f" not source '{self.projectB.source}' @@\n"
                )
                for projFileName in fileNamesOnlyInA:
                    writeFile.write(f">>>> {projFileName}\n")
                writeFile.write(
                    f"@@ Project files in source '{self.projectB.source}' but"
                    f" not source '{self.projectA.source}' @@\n"
                )
                for projFileName in fileNamesOnlyInB:
                    writeFile.write(f">>>> {projFileName}\n")
        for projFileName in fileNamesInAandB:
            compArr1 = []
            compArr2 = []
            for elem in self.projectA.files[projFileName].contents:
                compArr1.append(str(elem))
            for elem in self.projectB.files[projFileName].contents:
                compArr2.append(str(elem))
            compArr1.sort()
            compArr2.sort()

            doesDiffExist = False
            if (
                self.projectA.files[projFileName].contents
                != self.projectB.files[projFileName].contents
            ):
                doesDiffExist = True
            if doesDiffExist:
                os.makedirs(diffDirPath, exist_ok=True)
                projFileName_ = projFileName.split("/")[-1]
                diffFilePath = f"{diffDirPath}/{projFileName_}.diff"

                htmlDiff = HtmlDiff()
                difference = htmlDiff.make_file(
                    compArr1,
                    compArr2,
                    fromdesc="Source A",
                    todesc="Source B",
                    context=False,
                )
                diffFile = f"{diffDirPath}/{projFileName_}_diff.html"
                with open(diffFile, "w") as f:
                    for line in difference.splitlines():
                        print(line, file=f)

                with open(diffFilePath, "w+") as diffFile:
                    diffFile.writelines(
                        difflib.unified_diff(
                            self.projectA.files[projFileName].contents,
                            self.projectB.files[projFileName].contents,
                            fromfile=f"{self.projectA.source}: {projFileName}",
                            tofile=f"{self.projectB.source}: {projFileName}",
                            lineterm="\n",
                        )
                    )


class PPMACCompare(object):
    """
    Compare two PowerPMAC objects
    """

    def __init__(self, ppmacA, ppmacB, compareDir):
        self.ppmacInstanceA = ppmacA
        self.ppmacInstanceB = ppmacB
        self.compareDir = compareDir
        # Set of element names only in A
        self.elemNamesOnlyInA = {}
        # Set of element names only in B
        self.elemNamesOnlyInB = {}
        # Set of element names in both A and B
        self.elemNamesInAandB = {}
        # Dictionary of active elements only in A. Keys refer to the active
        # elem names, the values are PowerPMAC.activeElement objects.
        self.activeElemsOnlyInA = {}
        # Dictionary of active elements only in B. Keys refer to the active
        # elem names, the values are PowerPMAC.activeElement objects.
        self.activeElemsOnlyInB = {}
        # A nested dictionary. The outer keys refer to the active elem names,
        # the inner keys refer to ppmac instance (A or B), the values
        # are PowerPMAC.activeElement objects.
        self.activeElemsInAandB = {}
        self.progNamesOnlyInA = {}
        self.progNamesOnlyInB = {}
        self.progNamesInAandB = {}

    def setPPMACInstanceA(self, ppmacA):
        self.ppmacInstanceA = ppmacA

    def setPPMACInstanceB(self, ppmacB):
        self.ppmacInstanceB = ppmacB

    def compareActiveElements(self):
        elementNamesA = set(self.ppmacInstanceA.activeElements.keys())
        elementNamesB = set(self.ppmacInstanceB.activeElements.keys())
        self.elemNamesOnlyInA = elementNamesA - elementNamesB
        self.elemNamesOnlyInB = elementNamesB - elementNamesA
        self.elemNamesInAandB = elementNamesA & elementNamesB
        self.activeElemsOnlyInA = {
            elemName: self.ppmacInstanceA.activeElements[elemName]
            for elemName in self.elemNamesOnlyInA
        }
        self.activeElemsOnlyInB = {
            elemName: self.ppmacInstanceB.activeElements[elemName]
            for elemName in self.elemNamesOnlyInB
        }

        for elemName in self.elemNamesInAandB:
            self.activeElemsInAandB[elemName] = {
                "A": self.ppmacInstanceA.activeElements[elemName],
                "B": self.ppmacInstanceB.activeElements[elemName],
            }

        self.writeActiveElemDifferencesToFile()

    def comparePrograms(self):
        outputDir = f"{self.compareDir}/active/programs"
        programsA = mergeDicts(
            self.ppmacInstanceA.motionPrograms,
            self.ppmacInstanceA.subPrograms,
            self.ppmacInstanceA.plcPrograms,
            self.ppmacInstanceA.forwardPrograms,
            self.ppmacInstanceA.inversePrograms,
        )
        programsB = mergeDicts(
            self.ppmacInstanceB.motionPrograms,
            self.ppmacInstanceB.subPrograms,
            self.ppmacInstanceB.plcPrograms,
            self.ppmacInstanceB.forwardPrograms,
            self.ppmacInstanceB.inversePrograms,
        )
        progNamesA = set(programsA.keys())
        progNamesB = set(programsB.keys())
        self.progNamesOnlyInA = progNamesA - progNamesB
        self.progNamesOnlyInB = progNamesB - progNamesA
        self.progNamesInAandB = progNamesA & progNamesB
        if len(self.progNamesOnlyInA) or len(self.progNamesOnlyInB):
            os.makedirs(outputDir, exist_ok=True)

            compArr1 = []
            compArr2 = []
            for name in progNamesA:
                compArr1.append(name)
            for name in progNamesB:
                compArr2.append(name)
            compArr1.sort()
            compArr2.sort()

            htmlDiff = HtmlDiff()
            difference = htmlDiff.make_file(
                compArr1,
                compArr2,
                fromdesc="Programs in Source A",
                todesc="Programs in Source B",
                context=False,
            )
            diffFile = f"{outputDir}/MissingPrograms.html"
            with open(diffFile, "w") as f:
                for line in difference.splitlines():
                    print(line, file=f)

            with open(f"{outputDir}/missingPrograms.txt", "w+") as writeFile:
                writeFile.write(
                    f"@@ Programs in source '{self.ppmacInstanceA.source}' but"
                    f" not source '{self.ppmacInstanceB.source}'\n"
                )
                for progName in self.progNamesOnlyInA:
                    writeFile.write(f">>>> {progName}\n")
                    writeFile.write(f"{programsA[progName].printInfo()}\n")
                writeFile.write(
                    f"@@ Programs in source '{self.ppmacInstanceB.source}' but"
                    f" not source '{self.ppmacInstanceA.source}'\n"
                )
                for progName in self.progNamesOnlyInB:
                    writeFile.write(f">>>> {progName}\n")
                    writeFile.write(f"{programsB[progName].printInfo()}\n")
        for progName in self.progNamesInAandB:
            filePath = f"{outputDir}/{progName}.diff"

            compArr1 = []
            compArr2 = []
            for elem in programsA[progName].listing:
                compArr1.append(str(elem))
            for elem in programsB[progName].listing:
                compArr2.append(str(elem))
            compArr1.sort()
            compArr2.sort()

            doesDiffExist = False
            if programsA[progName].listing != programsB[progName].listing:
                doesDiffExist = True
            if doesDiffExist:
                os.makedirs(outputDir, exist_ok=True)
                htmlDiff = HtmlDiff()
                difference = htmlDiff.make_file(
                    compArr1,
                    compArr2,
                    fromdesc="Source A",
                    todesc="Source B",
                    context=False,
                )
                diffFile = f"{outputDir}/{progName}_diff.html"
                with open(diffFile, "w") as f:
                    for line in difference.splitlines():
                        print(line, file=f)
                with open(filePath, "w+") as writeFile:
                    writeFile.writelines(
                        difflib.unified_diff(
                            programsA[progName].listing,
                            programsB[progName].listing,
                            fromfile=f"{self.ppmacInstanceA.source}: {progName}",
                            tofile=f"{self.ppmacInstanceB.source}: {progName}",
                            lineterm="\n",
                        )
                    )

    def compareCoordSystemAxesDefinitions(self):
        outputDir = f"{self.compareDir}/active/axes"
        coordSysDefsA = self.ppmacInstanceA.coordSystemDefs
        coordSysDefsB = self.ppmacInstanceB.coordSystemDefs
        coordSysNumbersA = set(coordSysDefsA.keys())
        coordSysNumbersB = set(coordSysDefsB.keys())
        self.coordSystemsOnlyInA = coordSysNumbersA - coordSysNumbersB
        self.coordSystemsOnlyInB = coordSysNumbersB - coordSysNumbersA
        self.coordSystemsInAandB = coordSysNumbersA & coordSysNumbersB

        if len(self.coordSystemsOnlyInA) or len(self.coordSystemsOnlyInB):
            os.makedirs(outputDir, exist_ok=True)

            compArr1 = []
            compArr2 = []
            for elemName in coordSysNumbersA:
                compArr1.append("&" + str(elemName))
            for elemName in coordSysNumbersB:
                compArr2.append("&" + str(elemName))
            compArr1.sort()
            compArr2.sort()
            htmlDiff = HtmlDiff()
            difference = htmlDiff.make_file(
                compArr1,
                compArr2,
                fromdesc="Coord Systems defined in A",
                todesc="Coord Systems defined in B",
                context=False,
            )
            diffFile = f"{outputDir}/MissingCoordSystems.html"
            with open(diffFile, "w") as f:
                for line in difference.splitlines():
                    print(line, file=f)

            with open(f"{outputDir}/missingCoordSystems.txt", "w+") as writeFile:
                writeFile.write(
                    "@@ Coordinate Systems defined in source"
                    f" '{self.ppmacInstanceA.source}' but not source"
                    f" '{self.ppmacInstanceB.source}'\n"
                )
                for coordSysNumber in self.coordSystemsOnlyInA:
                    writeFile.write(f">>>> &{coordSysNumber}\n")
                    writeFile.write(
                        self.ppmacInstanceA.coordSystemDefs[coordSysNumber].printInfo()
                    )
                writeFile.write(
                    "@@ Coordinate Systems defined in source"
                    f" '{self.ppmacInstanceB.source}' but not source"
                    f" '{self.ppmacInstanceA.source}'\n"
                )
                for coordSysNumber in self.coordSystemsOnlyInB:
                    writeFile.write(f">>>> &{coordSysNumber}\n")
                    writeFile.write(
                        self.ppmacInstanceB.coordSystemDefs[coordSysNumber].printInfo()
                    )
        for coordSysNumber in self.coordSystemsInAandB:
            filePath = f"{outputDir}/cs{coordSysNumber}Axes.diff"
            added, removed, modified, same = comparedicts(
                coordSysDefsA[coordSysNumber].motor,
                coordSysDefsB[coordSysNumber].motor,
            )
            if len(modified) > 0:
                os.makedirs(outputDir, exist_ok=True)

                compArr1 = []
                compArr2 = []
                compArr1.append(
                    str(self.ppmacInstanceA.coordSystemDefs[coordSysNumber].printInfo())
                )
                compArr2.append(
                    str(self.ppmacInstanceB.coordSystemDefs[coordSysNumber].printInfo())
                )
                compArr1.sort()
                compArr2.sort()
                htmlDiff = HtmlDiff()
                difference = htmlDiff.make_file(
                    compArr1,
                    compArr2,
                    fromdesc="Coord Systems defined in A",
                    todesc="Coord Systems defined in B",
                    context=False,
                )
                diffFile = f"{outputDir}/cs{coordSysNumber}Axes_diff.html"
                with open(diffFile, "w") as f:
                    for line in difference.splitlines():
                        print(line, file=f)

                with open(filePath, "w+") as writeFile:
                    for motorNumber in modified:
                        writeFile.write(
                            "@@ Motor"
                            f" {motorNumber} @@\n{self.ppmacInstanceA.source}"
                            f" definition = "
                            f"&{coordSysNumber}#{motorNumber}"
                            f"->{modified[motorNumber][0]}\n"
                            f"{self.ppmacInstanceB.source} definition = "
                            f"&{coordSysNumber}#{motorNumber}"
                            f"->{modified[motorNumber][1]}\n"
                        )

    def writeActiveElemDifferencesToFile(self):
        outputDir = f"{self.compareDir}/active"
        os.makedirs(outputDir, exist_ok=True)
        # Create diff file with all differences
        compArr1 = []
        compArr2 = []
        for elemName in self.ppmacInstanceA.activeElements.keys():
            compArr1.append(
                str(self.ppmacInstanceA.activeElements[elemName].printInfo())
            )
        for elemName in self.ppmacInstanceB.activeElements.keys():
            compArr2.append(
                str(self.ppmacInstanceB.activeElements[elemName].printInfo())
            )
        compArr1.sort()
        compArr2.sort()

        htmlDiff = HtmlDiff()
        difference = htmlDiff.make_file(
            compArr1, compArr2, fromdesc="Source A", todesc="Source B", context=True
        )
        diffFile = f"{outputDir}/ActiveElements_diff.html"
        with open(diffFile, "w") as f:
            for line in difference.splitlines():
                print(line, file=f)

        diffFiles = {}
        try:
            sourceA = self.ppmacInstanceA.source
            sourceB = self.ppmacInstanceB.source
            # write to file elements that are in ppmacA but not ppmacB
            for elemName in self.elemNamesOnlyInA:
                file = (
                    f"{self.compareDir}/active/"
                    f"{self.activeElemsOnlyInA[elemName].category}.diff"
                )
                if file not in diffFiles:
                    diffFiles[file] = open(file, "w+")
                diffFiles[file].write(
                    f"@@ Active elements in source '{sourceA}' but not source"
                    f" '{sourceB}' @@\n"
                )
                diffFiles[file].write(
                    ">>>> "
                    + self.activeElemsOnlyInA[elemName].name
                    + " = "
                    + self.activeElemsOnlyInA[elemName].value
                    + "\n"
                )
            # write to file elements that are in ppmacB but not ppmacA
            for elemName in self.elemNamesOnlyInB:
                file = (
                    f"{self.compareDir}/active/"
                    f"{self.activeElemsOnlyInB[elemName].category}.diff"
                )
                if file not in diffFiles:
                    print("open file " + file)
                    diffFiles[file] = open(file, "w+")
                diffFiles[file].write(
                    f"@@ Active elements in source '{sourceB}' but not source"
                    f" '{sourceA}' @@\n"
                )
                diffFiles[file].write(
                    ">>>> "
                    + self.activeElemsOnlyInB[elemName].name
                    + " = "
                    + self.activeElemsOnlyInB[elemName].value
                    + "\n"
                )
            # write to file elements that in ppmacB but not ppmacA but whose
            # values differ
            for elemName in self.elemNamesInAandB:
                valA = self.activeElemsInAandB[elemName]["A"].value
                valB = self.activeElemsInAandB[elemName]["B"].value
                if valA != valB:
                    file = (
                        f"{self.compareDir}/active/"
                        f"{self.ppmacInstanceA.activeElements[elemName].category}.diff"
                    )
                    if file not in diffFiles:
                        print("open file " + file)
                        diffFiles[file] = open(file, "w+")
                    diffFiles[file].write(
                        f"@@ Active elements in source '{sourceA}' and source"
                        " '{sourceB}' with different values @@\n"
                    )
                    diffFiles[file].write(
                        f">>>> {elemName} @@\n{self.ppmacInstanceA.source} value ="
                        f" {valA}\n{self.ppmacInstanceB.source} value = {valB}\n"
                    )
        finally:
            for file in diffFiles:
                diffFiles[file].close()


class PPMACRepositoryWriteRead(object):
    def __init__(self, ppmac, repositoryPath):
        self.ppmacInstance = ppmac
        # source - will be set somewhere else
        if self.ppmacInstance.source == "unknown":
            self.ppmacInstance.source = "repository"
        self.repositoryPath = repositoryPath

    def setPPMACInstance(self, ppmac):
        self.ppmacInstance = ppmac
        if self.ppmacInstance.source == "unknown":
            self.ppmacInstance.source = "repository"

    def setRepositoryPath(self, path):
        self.repositoryPath = path

    def writeActiveState(self):
        os.makedirs(self.repositoryPath + "/active", exist_ok=True)
        self.writeDataStructures()
        self.writeActiveElements()
        self.writeAllPrograms()
        self.writeCSAxesDefinitions()

    def writeDataStructures(self):
        file = self.repositoryPath + "/active/dataStructures.txt"
        with open(file, "w+") as writeFile:
            for dataStructure in self.ppmacInstance.dataStructures:
                writeFile.write(
                    self.ppmacInstance.dataStructures[dataStructure].printInfo() + "\n"
                )

    def writeActiveElements(self):
        file = self.repositoryPath + "/active/activeElements.txt"
        with open(file, "w+") as writeFile:
            for elem in self.ppmacInstance.activeElements:
                writeFile.write(
                    self.ppmacInstance.activeElements[elem].printInfo() + "\n"
                )

    def writePrograms(self, ppmacProgs, progsDir):
        """
        Write contents of dictionary of PPMAC programs to files. Each dict key
        corresponds to separate file written.
        :param ppmacProgs: dictionary of programs.
        :return:
        """
        progNames = ppmacProgs.keys()
        for progName in progNames:
            fileName = progsDir + "/" + progName.replace("&", "CS")
            with open(fileName, "w+") as writeFile:
                writeFile.write(ppmacProgs[progName].printInfo())
                for line in ppmacProgs[progName].listing:
                    writeFile.write(line)

    def writeAllPrograms(self):
        progsDir = self.repositoryPath + "/active/programs"
        os.makedirs(progsDir, exist_ok=True)
        self.writePrograms(self.ppmacInstance.motionPrograms, progsDir)
        self.writePrograms(self.ppmacInstance.subPrograms, progsDir)
        self.writePrograms(self.ppmacInstance.plcPrograms, progsDir)
        self.writePrograms(self.ppmacInstance.forwardPrograms, progsDir)
        self.writePrograms(self.ppmacInstance.inversePrograms, progsDir)

    def writeCSAxesDefinitions(self):
        csAxesDefsPath = self.repositoryPath + "/active/axes"
        os.makedirs(csAxesDefsPath, exist_ok=True)
        for coordSystem in self.ppmacInstance.coordSystemDefs:
            fileName = f"{csAxesDefsPath}/cs{coordSystem}Axes"
            with open(fileName, "w+") as writeFile:
                writeFile.write(
                    self.ppmacInstance.coordSystemDefs[coordSystem].printInfo()
                )

    def readAndStoreActiveElements(self):
        fileName = self.repositoryPath + "/active/activeElements.txt"
        with open(fileName, "r") as readFile:
            for line in readFile:
                line = line.split()
                line = [item.strip() for item in line]
                if line == "":
                    continue
                key = line[0]
                value = line[
                    0:
                ]  # need to deal with the list of indices which is tha last column(s)
                self.ppmacInstance.activeElements[
                    key
                ] = self.ppmacInstance.ActiveElement(*value[0:5])

    def readAndStoreCSAxesDefinitions(self):
        csAxesDefsPath = self.repositoryPath + "/active/axes"
        csAxesFileNames = [
            file
            for file in os.listdir(csAxesDefsPath)
            if os.path.isfile(f"{csAxesDefsPath}/{file}")
        ]
        for fileName in csAxesFileNames:
            file = f"{csAxesDefsPath}/{fileName}"
            csNumber = str(re.search(r"\d+", fileName).group())
            self.ppmacInstance.coordSystemDefs[csNumber] = [csNumber, []]
            with open(file, "r") as readFile:
                motorDefs = []
                for line in readFile:
                    line = line.strip()
                    if line == "":
                        continue
                    csNumber_ = PPMACLexer(line).pop(1)[1]
                    if csNumber is not csNumber_:
                        raise IOError(
                            f"Inconsistent coordinate system numbers in file: {file}"
                        )
                    motorDefinition = line.rstrip("\n")
                    motorDefs.append(motorDefinition)
            self.ppmacInstance.coordSystemDefs[
                csNumber
            ] = self.ppmacInstance.CoordSystemDefinition(csNumber, motorDefs)

    def readAndStoreBufferedPrograms(self):
        progsPath = self.repositoryPath + "/active/programs"
        progFileNames = [
            file
            for file in os.listdir(progsPath)
            if os.path.isfile(f"{progsPath}/{file}")
        ]
        for fileName in progFileNames:
            fileName = f"{progsPath}/{fileName}"
            with open(fileName, "r") as readFile:
                header = readFile.readline().split()
                header = [item.strip(",") for item in header]
                progType, progName, progSize, progOffset = (
                    header[i] for i in [0, 1, 3, 5]
                )
                progListing = []
                for line in readFile:
                    if line.strip() == "":
                        continue
                    progListing.append(line)
            if progType == "Motion":
                self.ppmacInstance.motionPrograms[
                    progName
                ] = self.ppmacInstance.Program(
                    progName, progOffset, progSize, progType, progListing
                )
            elif progType == "SubProg":
                self.ppmacInstance.subPrograms[progName] = self.ppmacInstance.Program(
                    progName, progOffset, progSize, progType, progListing
                )
            elif progType == "Plc":
                self.ppmacInstance.plcPrograms[progName] = self.ppmacInstance.Program(
                    progName, progOffset, progSize, progType, progListing
                )
            elif progType == "Forward":
                progCoordSystem = header[7]
                self.ppmacInstance.forwardPrograms[
                    progName
                ] = self.ppmacInstance.KinematicTransform(
                    progName,
                    progSize,
                    progOffset,
                    progType,
                    progCoordSystem,
                    progListing,
                )
            elif progType == "Inverse":
                progCoordSystem = header[7]
                self.ppmacInstance.inversePrograms[
                    progName
                ] = self.ppmacInstance.KinematicTransform(
                    progName,
                    progSize,
                    progOffset,
                    progType,
                    progCoordSystem,
                    progListing,
                )


class PPMACHardwareWriteRead(object):
    def __init__(self, ppmac=None, tempDir=None):
        self.ppmacInstance = ppmac
        # Set PPMAC source to hardware
        if self.ppmacInstance.source == "unknown":
            self.ppmacInstance.source = "hardware"
        # Path to directory containing symbols tables files on local
        if tempDir is not None:
            self.local_db_path = tempDir
        # File containing list of all base Data Structures
        self.pp_swtbl0_txtfile = "pp_swtbl0.txt"
        # Standard Data Structure symbols tables
        self.pp_swtlbs_symfiles = ["pp_swtbl1.sym", "pp_swtbl2.sym", "pp_swtbl3.sym"]
        # Read maximum values for the various configuration parameters
        # self.readSysMaxes()

    def setPPMACInstance(self, ppmac):
        self.ppmacInstance = ppmac
        if self.ppmacInstance.source == "unknown":
            self.ppmacInstance.source = "hardware"

    def getCommandReturnInt(self, cmd):
        (cmdReturn, status) = sshClient.sendCommand(cmd)
        if status:
            cmdReturnInt = int(cmdReturn[0:])
        else:
            raise IOError(
                "Cannot retrieve variable value: error communicating with PMAC"
            )
        return cmdReturnInt

    def readSysMaxes(self):
        if self.ppmacInstance is None:
            raise RuntimeError("No Power PMAC object has been specified")
        # number of active motors
        self.ppmacInstance.numberOfMotors = self.getCommandReturnInt("Sys.MaxMotors")
        # number of active coordinate systems
        self.ppmacInstance.numberOfCoordSystems = self.getCommandReturnInt(
            "Sys.MaxCoords"
        )
        # number of active compensation tables.
        self.ppmacInstance.numberOfCompTables = self.getCommandReturnInt(
            "Sys.CompEnable"
        )
        # number of active cam tables
        self.ppmacInstance.numberOfCamTables = self.getCommandReturnInt("Sys.CamEnable")
        # number EtherCAT networks that can be enabled
        self.ppmacInstance.numberOfECATs = self.getCommandReturnInt("Sys.MaxEcats")
        # Number of available encoder tables
        self.ppmacInstance.numberOfEncTables = self.getCommandReturnInt(
            "Sys.MaxEncoders"
        )

    def sendCommand(self, cmd):
        (data, status) = sshClient.sendCommand(cmd)
        if not status:
            raise IOError(
                "Cannot retrieve data structure: error communicating with PMAC"
            )
        else:
            data = data.split("\r")[:-1]
        return data

    def swtblFileToList(self, pp_swtbl_file):
        """
        Generate a list of symbols from a symbols table file.
        :param pp_swtbl_file: full path to symbols table file.
        :return: swtbl_DSs: list of symbols, where each 'symbol' is represented by
        the contents of one row of the symbols table file.
        """
        symbols = []
        try:
            file = open(file=pp_swtbl_file, mode="r", encoding="ISO-8859-1")
            multi_line = ""
            for line in file:
                multi_line_ = multi_line
                if line[-2:] == "\\\n":
                    multi_line += line
                else:
                    line = multi_line_ + line
                    symbol = line.split("\x01")
                    symbol = [col.strip() for col in symbol]
                    multi_line = ""
                    symbols.append(symbol)
            file.close()
        except IOError as e:
            print(e)
        return symbols

    def getDataStructureCategory(self, dataStructure):
        if dataStructure.find(".") == -1:
            dataStructureCategory = dataStructure
        else:
            dataStructureCategory = dataStructure[0 : dataStructure.find(".")]
        return dataStructureCategory.replace("[]", "")

    def ignoreDataStructure(self, substructure, elementsToIgnore):
        """
        Determine whether a data structure or substructure should be ignored by
        checking if it or any of its parent, grandparent, great-grandparent etc.
        structures are included in the ignore list.
        :param substructure: name of data structure or substructure to check
        :return: True if data structure or substructure should be ignored, False
        otherwise
        """
        n = substructure.count(".")
        for _ in range(n + 1):
            if substructure in elementsToIgnore:
                return True
            substructure = substructure[0 : substructure.rfind(".")]
        return False

    def fillDataStructureIndices_i(
        self, dataStructure, activeElements, elementsToIgnore, timeout=None
    ):
        """
        Incrementally increase the index of a singly-indexed data structure and send
        the resulting command to the ppmac until the maximum accepted index is reached.
        Add the command string and return value of all commands accepted by
        the ppmac to the dictionary of active elements.
        :param dataStructure: String containing the data structure name.
        :param activeElements: Dictionary containing the current set of active elements,
        where the key is the element name, and the value is a tuple containing
        the return value from the ppmac and the active element name.
        :param elementsToIgnore: Set of data structures not to be added to
        activeElements.
        :return:
        """
        applyTimeout = False
        if isinstance(timeout, int) or isinstance(timeout, float):
            applyTimeout = True
            startTime = time.time()
        dataStructureCategory = self.getDataStructureCategory(dataStructure)
        i = 0
        cmd_accepted = True
        while cmd_accepted:
            i_idex_string = f"[{i}]"
            dataStructure_i = dataStructure.replace("[]", i_idex_string)
            if self.ignoreDataStructure(
                dataStructure_i.replace(i_idex_string, f"[{i}:]"), elementsToIgnore
            ):
                break
            if self.ignoreDataStructure(dataStructure_i, elementsToIgnore):
                i += 1
                continue
            # print(dataStructure_i)
            cmd_return = self.sendCommand(dataStructure_i)
            # print(cmd_return)
            if "ILLEGAL" in cmd_return[0]:
                cmd_accepted = False
            else:
                activeElements[dataStructure_i] = (
                    dataStructure_i,
                    cmd_return[0],
                    dataStructureCategory,
                    dataStructure,
                    [i],
                )
            if applyTimeout and time.time() - startTime > timeout:
                logging.info(
                    f"Timed-out generating active elements for {dataStructure}. "
                    f"Last i = {i}."
                )
                return
            i += 1

    def fillDataStructureIndices_ij(
        self, dataStructure, activeElements, elementsToIgnore, timeout=None
    ):
        """
        Incrementally increase the indices of a doubly-indexed data structure and send
        the resulting command to the ppmac until the maximum accepted indices are
        reached. Add the command string and return value of all commands accepted by
        the ppmac to the dictionary of active elements.
        :param dataStructure: String containing the data structure name.
        :param activeElements: Dictionary containing the current set of active
        elements, where the key is the element name, and the value is a tuple
        containing the return value from the ppmac and the active element name.
        :param elementsToIgnore: Set of data structures not to be added to
        activeElements.
        :return:
        """
        applyTimeout = False
        if isinstance(timeout, int) or isinstance(timeout, float):
            applyTimeout = True
            startTime = time.time()
        dataStructureCategory = self.getDataStructureCategory(dataStructure)
        i = 0
        last_i_accepted = i
        cmd_accepted = True
        while cmd_accepted:
            j = 0
            i_idex_string = f"[{i}]"
            dataStructure_i = nthRepl(dataStructure, "[]", i_idex_string, 1)
            if self.ignoreDataStructure(
                dataStructure_i.replace(i_idex_string, f"[{i}:]"), elementsToIgnore
            ):
                break
            if self.ignoreDataStructure(dataStructure_i, elementsToIgnore):
                last_i_accepted = i
                i += 1
                continue
            while cmd_accepted:
                j_idex_string = f"[{j}]"
                dataStructure_ij = nthRepl(dataStructure_i, "[]", j_idex_string, 1)
                if self.ignoreDataStructure(
                    dataStructure_ij.replace(f"[{i}]", "[]", 1).replace(
                        j_idex_string, f"[{j}:]", 1
                    ),
                    elementsToIgnore,
                ):
                    break
                if self.ignoreDataStructure(
                    dataStructure_ij.replace(f"[{i}]", "[]", 1), elementsToIgnore
                ):
                    j += 1
                    continue
                # print(dataStructure_ij)
                cmd_return = self.sendCommand(dataStructure_ij)
                # print(cmd_return)
                if "ILLEGAL" in cmd_return[0]:
                    cmd_accepted = False
                else:
                    last_i_accepted = i
                    activeElements[dataStructure_ij] = (
                        dataStructure_ij,
                        cmd_return[0],
                        dataStructureCategory,
                        dataStructure,
                        [i, j],
                    )
                if applyTimeout and time.time() - startTime > timeout:
                    logging.info(
                        f"Timed-out generating active elements for {dataStructure}. "
                        f"Last i,j = {i},{j}."
                    )
                    return
                j += 1
            cmd_accepted = True
            # print(last_i_accepted, i)
            if i - last_i_accepted > 1:
                cmd_accepted = False
            # print(cmd_accepted)
            i += 1
        # print(activeElements)

    def fillDataStructureIndices_ijk(
        self, dataStructure, activeElements, elementsToIgnore, timeout=None
    ):
        """
        Incrementally increase the indices of a triply-indexed data structure and send
        the resulting command to the ppmac until the maximum accepted indices are
        reached. Add the command string and return value of all commands accepted by
        the ppmac to the dictionary of active elements.
        :param dataStructure: String containing the data structure name.
        :param activeElements: Dictionary containing the current set of active
        elements, where the key is the element name, and the value is a
        tuple containing the return value from the ppmac and the active element name.
        :param elementsToIgnore: Set of data structures not to be added to
        activeElements.
        :return:
        """
        applyTimeout = False
        if isinstance(timeout, int) or isinstance(timeout, float):
            applyTimeout = True
            startTime = time.time()
        dataStructureCategory = self.getDataStructureCategory(dataStructure)
        i = 0
        last_i_accepted = i
        cmd_accepted = True
        # print(dataStructure)
        while cmd_accepted:
            j = 0
            last_j_accepted = j
            i_idex_string = f"[{i}]"
            dataStructure_i = nthRepl(dataStructure, "[]", i_idex_string, 1)
            if self.ignoreDataStructure(
                dataStructure_i.replace(i_idex_string, f"[{i}:]"), elementsToIgnore
            ):
                break
            if self.ignoreDataStructure(dataStructure_i, elementsToIgnore):
                last_i_accepted = i
                i += 1
                continue
            #    break
            while cmd_accepted:
                k = 0
                j_idex_string = f"[{j}]"
                dataStructure_ij = nthRepl(dataStructure_i, "[]", j_idex_string, 1)
                if self.ignoreDataStructure(
                    dataStructure_ij.replace(f"[{i}]", "[]", 1).replace(
                        j_idex_string, f"[{j}:]", 1
                    ),
                    elementsToIgnore,
                ):
                    break
                if self.ignoreDataStructure(
                    dataStructure_ij.replace(f"[{i}]", "[]", 1), elementsToIgnore
                ):
                    last_j_accepted = j
                    j += 1
                    continue
                #    break
                while cmd_accepted:
                    k_idex_string = f"[{k}]"
                    dataStructure_ijk = nthRepl(
                        dataStructure_ij, "[]", k_idex_string, 1
                    )
                    if self.ignoreDataStructure(
                        dataStructure_ijk.replace(f"[{i}]", "[]", 1)
                        .replace(f"[{j}]", "[]", 1)
                        .replace(k_idex_string, f"[{k}:]", 1),
                        elementsToIgnore,
                    ):
                        break
                    if self.ignoreDataStructure(
                        dataStructure_ijk.replace(f"[{i}]", "[]", 1).replace(
                            f"[{j}]", "[]", 1
                        ),
                        elementsToIgnore,
                    ):
                        #    break
                        k += 1
                        continue
                    # print(dataStructure_ijk)
                    cmd_return = self.sendCommand(dataStructure_ijk)
                    # print(cmd_return)
                    if "ILLEGAL" in cmd_return[0]:
                        cmd_accepted = False
                    else:
                        last_j_accepted = j
                        last_i_accepted = i
                        activeElements[dataStructure_ijk] = (
                            dataStructure_ijk,
                            cmd_return[0],
                            dataStructureCategory,
                            dataStructure,
                            [i, j, k],
                        )
                    if applyTimeout and time.time() - startTime > timeout:
                        logging.info(
                            f"Timed-out generating active elements for {dataStructure}."
                            f" Last i,j,k = {i},{j},{k}."
                        )
                        return
                    k += 1
                cmd_accepted = True
                if j - last_j_accepted > 1:
                    cmd_accepted = False
                j += 1
            cmd_accepted = True
            if i - last_i_accepted > 1:
                cmd_accepted = False
            i += 1
        # print(activeElements)

    def fillDataStructureIndices_ijkl(
        self, dataStructure, activeElements, elementsToIgnore, timeout=None
    ):
        """
        Incrementally increase the indices of a quadruply-indexed data structure and
        send the resulting command to the ppmac until the maximum accepted
        indices are reached. Add the command string and return value of all commands
        accepted by the ppmac to the dictionary of active elements.
        :param dataStructure: String containing the data structure name.
        :param activeElements: Dictionary containing the current set of active
        elements, where the key is the element name, and the value is a tuple
        containing the return value from the ppmac and the active element name.
        :param elementsToIgnore: Set of data structures not to be added to
        activeElements.
        :return:
        """
        applyTimeout = False
        if isinstance(timeout, int) or isinstance(timeout, float):
            applyTimeout = True
            startTime = time.time()
        dataStructureCategory = self.getDataStructureCategory(dataStructure)
        i = 0
        last_i_accepted = i
        cmd_accepted = True
        while cmd_accepted:
            j = 0
            last_j_accepted = j
            i_idex_string = f"[{i}]"
            dataStructure_i = nthRepl(dataStructure, "[]", i_idex_string, 1)
            # if self.ignoreDataStructure(dataStructure_i, elementsToIgnore):
            #    break
            if self.ignoreDataStructure(
                dataStructure_i.replace(i_idex_string, f"[{i}:]"), elementsToIgnore
            ):
                break
            if self.ignoreDataStructure(dataStructure_i, elementsToIgnore):
                last_i_accepted = i
                i += 1
                continue
            while cmd_accepted:
                k = 0
                last_k_accepted = k
                j_idex_string = f"[{j}]"
                dataStructure_ij = nthRepl(dataStructure_i, "[]", j_idex_string, 1)
                if self.ignoreDataStructure(
                    dataStructure_ij.replace(f"[{i}]", "[]", 1).replace(
                        j_idex_string, f"[{j}:]", 1
                    ),
                    elementsToIgnore,
                ):
                    break
                if self.ignoreDataStructure(
                    dataStructure_ij.replace(f"[{i}]", "[]", 1), elementsToIgnore
                ):
                    last_j_accepted = j
                    j += 1
                    continue
                while cmd_accepted:
                    m = 0
                    k_idex_string = f"[{k}]"
                    dataStructure_ijk = nthRepl(
                        dataStructure_ij, "[]", k_idex_string, 1
                    )
                    if self.ignoreDataStructure(
                        dataStructure_ijk.replace(f"[{i}]", "[]", 1)
                        .replace(f"[{j}]", "[]", 1)
                        .replace(k_idex_string, f"[{k}:]", 1),
                        elementsToIgnore,
                    ):
                        break
                    if self.ignoreDataStructure(
                        dataStructure_ijk.replace(f"[{i}]", "[]", 1).replace(
                            f"[{j}]", "[]", 1
                        ),
                        elementsToIgnore,
                    ):
                        #    break
                        last_k_accepted = k
                        k += 1
                        continue
                    while cmd_accepted:
                        m_idex_string = f"[{m}]"
                        dataStructure_ijkm = nthRepl(
                            dataStructure_ijk, "[]", m_idex_string, 1
                        )
                        if self.ignoreDataStructure(
                            dataStructure_ijkm.replace(f"[{i}]", "[]", 1)
                            .replace(f"[{j}]", "[]", 1)
                            .replace(f"[{k}]", "[]", 1)
                            .replace(m_idex_string, f"[{m}:]", 1),
                            elementsToIgnore,
                        ):
                            break
                        if self.ignoreDataStructure(
                            dataStructure_ijkm.replace(f"[{i}]", "[]", 1)
                            .replace(f"[{j}]", "[]", 1)
                            .replace(f"[{k}]", "[]", 1),
                            elementsToIgnore,
                        ):
                            m += 1
                            continue
                        # print(dataStructure_ijkm)
                        cmd_return = self.sendCommand(dataStructure_ijkm)
                        # print(cmd_return)
                        if "ILLEGAL" in cmd_return[0]:
                            cmd_accepted = False
                        else:
                            last_k_accepted = k
                            last_j_accepted = j
                            last_i_accepted = i
                            activeElements[dataStructure_ijkm] = (
                                dataStructure_ijkm,
                                cmd_return[0],
                                dataStructureCategory,
                                dataStructure,
                                [i, j, k, m],
                            )
                        if applyTimeout and time.time() - startTime > timeout:
                            logging.info(
                                "Timed-out generating active elements for"
                                f" {dataStructure}. Last i,j,k,m = {i},{j},{k},{m}."
                            )
                            return
                        m += 1
                    cmd_accepted = True
                    if k - last_k_accepted > 1:
                        cmd_accepted = False
                    k += 1
                cmd_accepted = True
                if j - last_j_accepted > 1:
                    cmd_accepted = False
                j += 1
            cmd_accepted = True
            if i - last_i_accepted > 1:
                cmd_accepted = False
            i += 1
        # print(activeElements)

    def scpPPMACDatabaseToLocal(self, remote_db_path, local_db_path):
        if not os.path.isdir(local_db_path):
            os.system("mkdir " + local_db_path)
        scpFromPowerPMACtoLocal(
            source=remote_db_path, destination=local_db_path, recursive=True
        )

    def createDataStructuresFromSymbolsTables(self, pp_swtlbs_symfiles, local_db_path):
        """
        Read the symbols tables and create a list of data structure names contained
        within them.
        :return: dataStructures: list of data structure names
        """
        # Clear current data structure dictionary
        dataStructures = {}
        # swtbl0 = []
        # with open(self.local_db_path + '/' + self.pp_swtbl0_txtfile, 'r') as readFile:
        #    for line in readFile:
        #        swtbl0.append(line.replace('\n',''))
        pp_swtbls = []
        for pp_swtbl_file in pp_swtlbs_symfiles:
            pp_swtbls.append(self.swtblFileToList(local_db_path + "/" + pp_swtbl_file))
        swtbl1_nparray = np.asarray(pp_swtbls[0])
        swtbl2_nparray = np.asarray(pp_swtbls[1])
        swtbl3_nparray = np.asarray(pp_swtbls[2])
        # for baseDS in swtbl0:
        #    print(baseDS)
        #    substruct_01 = False
        for i in range(swtbl1_nparray.shape[0]):
            #    if baseDS == swtbl1_nparray[i, 1]:
            #        substruct_01 = True
            substruct_12 = False
            for j in range(swtbl2_nparray.shape[0]):
                if swtbl1_nparray[i, 2] == swtbl2_nparray[j, 1]:
                    if (
                        swtbl1_nparray[i, 1].replace("[]", "")
                        != swtbl2_nparray[j, 5].replace("[]", "")
                    ) and (swtbl2_nparray[j, 5] != "NULL"):
                        continue
                    substruct_12 = True
                    substruct_23 = False
                    for k in range(swtbl3_nparray.shape[0]):
                        if swtbl2_nparray[j, 2] == swtbl3_nparray[k, 1]:
                            if (
                                swtbl1_nparray[i, 1].replace("[]", "")
                                != swtbl3_nparray[k, 5].replace("[]", "")
                            ) and (swtbl3_nparray[k, 5] != "NULL"):
                                continue
                            substruct_23 = True
                            dsName = (
                                swtbl1_nparray[i, 1]
                                + "."
                                + swtbl2_nparray[j, 1]
                                + "."
                                + swtbl3_nparray[k, 1]
                                + "."
                                + swtbl3_nparray[k, 2]
                            )
                            # print(dsName)
                            dataStructures[dsName] = [
                                dsName,
                                swtbl1_nparray[i, 1],
                                *(swtbl3_nparray[k, 3:].tolist()),
                            ]
                    if substruct_23 is False:
                        dsName = (
                            swtbl1_nparray[i, 1]
                            + "."
                            + swtbl2_nparray[j, 1]
                            + "."
                            + swtbl2_nparray[j, 2]
                        )
                        # print(dsName)
                        dataStructures[dsName] = [
                            dsName,
                            swtbl1_nparray[i, 1],
                            *(swtbl2_nparray[j, 3:].tolist()),
                        ]
            if substruct_12 is False:
                dsName = swtbl1_nparray[i, 1] + "." + swtbl1_nparray[i, 2]
                # print(dsName)
                dataStructures[dsName] = [
                    dsName,
                    swtbl1_nparray[i, 1],
                    *(swtbl1_nparray[i, 3:].tolist()),
                ]
        # if substruct_01 == False:
        #    print(baseDS)
        #    dataStructures.append(baseDS)
        #    writeFile.write(baseDS.printInfo() + '\n')
        return dataStructures

    def checkDataStructuresValidity(self, dataStructures):
        """
        Remove invalid data structures from a dictionary of data structures. A data
        structure is defined as invalid if the ppmac rejects it when its indices
        are filled-in as zero.
        :param dataStructures:
        :return:
        """
        logging.info(
            "checkDataStructuresValidity: checking if all data structures are valid, "
            "and removing any invalid data structures..."
        )
        invalidCount = 0
        for ds in list(dataStructures):
            cmd_return = self.sendCommand(ds.replace("[]", "[0]"))
            if "ILLEGAL" in cmd_return[0]:
                logging.debug(
                    f"{ds.replace('[]', '[0]')} not a valid ppmac command, deleting"
                    " from dictionary of data structures."
                )
                del dataStructures[ds]
                invalidCount += 1
        logging.info(
            f"checkDataStructuresValidity: removed {invalidCount} invalid data"
            " structures."
        )
        return dataStructures

    def getActiveElementsFromDataStructures(
        self, dataStructures, elementsToIgnore, recordTimings=False, timeout=None
    ):
        """
        Generate a dictionary of active elements from an iterable containing data
        structure names.
        :param dataStructures: Iterable containing data structure names
        :param elementsToIgnore: Set of data structures not to be added to included
        in the active elements read from the ppmac.
        :return: Dictionary containing the current set of active elements, where the
        key is the active element name, and the value is a tuple containing the return
        value from the ppmac and the active element name.
        """
        logging.info(
            "getActiveElementsFromDataStructures: generating dictionary of active"
            " elements..."
        )
        fncStartTime = time.time()
        activeElements = {}
        for ds in dataStructures:
            loopStartTime = time.time()
            N_brackets = ds.count("[]")
            if N_brackets == 0:
                value = self.sendCommand(ds)[0]
                category = self.getDataStructureCategory(ds)
                activeElements[ds] = (ds, value, category, ds, None)
            elif N_brackets == 1:
                self.fillDataStructureIndices_i(
                    ds, activeElements, elementsToIgnore, timeout=timeout
                )
            elif N_brackets == 2:
                self.fillDataStructureIndices_ij(
                    ds, activeElements, elementsToIgnore, timeout=timeout
                )
            elif N_brackets == 3:
                self.fillDataStructureIndices_ijk(
                    ds, activeElements, elementsToIgnore, timeout=timeout
                )
            elif N_brackets == 4:
                self.fillDataStructureIndices_ijkl(
                    ds, activeElements, elementsToIgnore, timeout=timeout
                )
            else:
                logging.info(
                    f"Too many indexed substructures in data structure '{ds}'."
                    " Ignoring."
                )
                continue
            if recordTimings:
                logging.info(ds + f"   time: {time.time() - loopStartTime} sec")
        logging.info("Finished generating dictionary of active elements. ")
        logging.info(f"Total time = {time.time() - fncStartTime} sec")
        return activeElements

    def expandSplicedIndices(self, splicedDataStructure):
        """
        Stuff
        :param splicedDataStructure: String containing a data structure name with one
        filled index that may or may not be spliced to indicate a range of values
        :return:
        """
        if splicedDataStructure.count(":") > 1:
            raise ("Too many indices")
        elif ":" not in splicedDataStructure:
            return [splicedDataStructure]
        else:
            splicedIndices = re.search("\\[([0-9]+):([0-9]+)\\]", splicedDataStructure)
            if splicedIndices is None:
                return [splicedDataStructure]
            startIndex = int(splicedIndices.group(1))
            endIndex = int(splicedIndices.group(2))
            expandedDataStructure = [
                re.sub("([0-9]+:[0-9]+)", str(i), splicedDataStructure)
                for i in range(startIndex, endIndex + 1)
            ]
            return expandedDataStructure

    def generateIgnoreSet(self, ignoreFile):
        ignore = []
        logging.info(f"Using ignore file {ignoreFile}.")
        with open(ignoreFile, "r") as readFile:
            for line in readFile:
                line = line.split("#", 1)[0]
                line = line.strip()
                ignore += line.split()
        expandedIgnore = [
            item
            for dataStructure in ignore
            for item in self.expandSplicedIndices(dataStructure)
        ]
        return set(expandedIgnore)

    def copyDict(self, destValueType, sourceDict):
        return {key: destValueType(*value) for key, value in sourceDict.items()}

    @timer
    def readAndStoreActiveState(self, pathToIgnoreFile):
        if self.local_db_path is None:
            raise IOError(
                "Need to specify temporary directory where Power PMAC Database can be"
                " copied to."
            )
        os.makedirs(self.local_db_path, exist_ok=True)
        self.scpPPMACDatabaseToLocal("/var/ftp/usrflash/Database/*", self.local_db_path)
        # Store data structures in ppmac object
        dataStructures = self.createDataStructuresFromSymbolsTables(
            self.pp_swtlbs_symfiles, self.local_db_path
        )
        validDataStructures = self.checkDataStructuresValidity(dataStructures)
        # Store the dictionary of data strutures in the ppmac object
        self.ppmacInstance.dataStructures = self.copyDict(
            self.ppmacInstance.DataStructure, validDataStructures
        )
        # Store active elements in ppmac object
        elementsToIgnore = self.generateIgnoreSet(pathToIgnoreFile)
        activeElements = self.getActiveElementsFromDataStructures(
            validDataStructures, elementsToIgnore, recordTimings=True
        )  # timeout=10.0
        self.ppmacInstance.activeElements = self.copyDict(
            self.ppmacInstance.ActiveElement, activeElements
        )
        bufferedProgramsInfo = self.getBufferedProgramsInfo()
        self.appendBufferedProgramsInfoWithListings(bufferedProgramsInfo)
        self.ppmacInstance.forwardPrograms = self.copyDict(
            self.ppmacInstance.KinematicTransform, bufferedProgramsInfo["Forward"]
        )
        self.ppmacInstance.inversePrograms = self.copyDict(
            self.ppmacInstance.KinematicTransform, bufferedProgramsInfo["Inverse"]
        )
        self.ppmacInstance.subPrograms = self.copyDict(
            self.ppmacInstance.Program, bufferedProgramsInfo["SubProg"]
        )
        self.ppmacInstance.motionPrograms = self.copyDict(
            self.ppmacInstance.Program, bufferedProgramsInfo["Motion"]
        )
        self.ppmacInstance.plcPrograms = self.copyDict(
            self.ppmacInstance.Program, bufferedProgramsInfo["Plc"]
        )
        coordSystemMotorDefs = self.getCoordSystemMotorDefinitions()
        self.ppmacInstance.coordSystemDefs = self.copyDict(
            self.ppmacInstance.CoordSystemDefinition, coordSystemMotorDefs
        )

    def getCoordSystemMotorDefinitions(self):
        coordSystemMotorDefinitions = {}
        currentCoordSystem = self.sendCommand("&")
        axesDefStatements = self.sendCommand("#*->")
        for definition in axesDefStatements:
            motorDefinitionTokens = PPMACLexer(definition)
            firstToken = motorDefinitionTokens.pop()[1]  # either '&' or '#' token
            if firstToken == "&":
                currentCoordSystem = motorDefinitionTokens.pop()[
                    1
                ]  # coord system number token
                motorDefinitionTokens.pop()[1]  # '#' token
            elif firstToken != "#":
                raise IOError(f"Unexpected token '{firstToken}'")
            motorDefinition = (
                f"&{currentCoordSystem}#{motorDefinitionTokens.getTokensAsString()}"
            )
            if currentCoordSystem in coordSystemMotorDefinitions:
                coordSystemMotorDefinitions[currentCoordSystem][1].append(
                    motorDefinition
                )
            else:
                coordSystemMotorDefinitions[currentCoordSystem] = [
                    currentCoordSystem,
                    [motorDefinition],
                ]
        return coordSystemMotorDefinitions

    def appendBufferedProgramsInfoWithListings(self, bufferedProgramsInfo):
        for programType in bufferedProgramsInfo.keys():
            for programInfo in bufferedProgramsInfo[programType].values():
                programName = programInfo[0]
                programListing = []
                if programType == "Forward" or programType == "Inverse":
                    progCoordSystem = programInfo[4]
                    listing = self.sendCommand(f"&{progCoordSystem} list {programType}")
                else:
                    listing = self.sendCommand(f"list {programName}")
                for line in listing:
                    programListing.append(line + "\n")
                programInfo.append(programListing)

    def getBufferedProgramsInfo(self):
        motion, subProgs, plcs, inverse, forward = ({} for _ in range(5))
        programBuffers = self.sendCommand("buffer")
        for progBuffInfo in programBuffers:
            progBuffInfo = progBuffInfo.split()
            if len(progBuffInfo) < 5:
                continue
            progName = progBuffInfo[0].strip("., ")
            progOffset = progBuffInfo[2].strip("., ")
            progSize = progBuffInfo[4].strip("., ")
            if "SubProg" in progName:
                subProgs[progName] = [progName, progOffset, progSize, "SubProg"]
            elif "Prog" in progName:
                motion[progName] = [progName, progOffset, progSize, "Motion"]
            elif "Plc" in progName:
                plcs[progName] = [progName, progOffset, progSize, "Plc"]
            elif "Inverse" in progName:
                progCoordSystem = progName.rstrip("Inverse").lstrip("&")
                inverse[progName] = [
                    progName,
                    progOffset,
                    progSize,
                    "Inverse",
                    progCoordSystem,
                ]
            elif "Forward" in progName:
                progCoordSystem = progName.rstrip("Forward").lstrip("&")
                forward[progName] = [
                    progName,
                    progOffset,
                    progSize,
                    "Forward",
                    progCoordSystem,
                ]
        return {
            "SubProg": subProgs,
            "Motion": motion,
            "Plc": plcs,
            "Inverse": inverse,
            "Forward": forward,
        }


class PowerPMAC:
    class DataStructure:
        def __init__(
            self,
            name="",
            base="",
            field1="",
            field2="",
            field3="",
            field4="",
            field5="",
            field6="",
            field7="",
            field8="",
            field9="",
            field10="",
            field11="",
            field12="",
        ):
            self.name = name
            self.base = base
            self.field1 = field1
            self.field2 = field2
            self.field3 = field3
            self.field4 = field4
            self.field5 = field5
            self.field6 = field6
            self.field7 = field7
            self.field8 = field8
            self.field9 = field9
            self.field10 = field10
            self.field11 = field11
            self.field12 = field12

        def printInfo(self):
            s = (
                self.name
                + ", "
                + self.base
                + ", "
                + self.field1
                + ", "
                + self.field2
                + ", "
                + self.field3
                + ", "
                + self.field4
                + ", "
                + self.field5
                + ", "
                + self.field6
                + ", "
                + self.field7
                + ", "
                + self.field8
                + ", "
                + self.field9
                + ", "
                + self.field10
                + ", "
                + self.field11
                + ", "
                + self.field12
            )
            return s

    class ActiveElement:
        def __init__(
            self, name="", value="", category="", dataStructure="", indices=[]
        ):
            self.name = name
            self.value = value
            if dataStructure == "":
                self.dataStructure = re.sub("\\[([0-9]+)\\]", "[]", self.name)
            else:
                self.dataStructure = dataStructure
            if category == "":
                if self.dataStructure.find(".") == -1:
                    category = self.dataStructure
                else:
                    category = self.dataStructure[0 : self.dataStructure.find(".")]
                self.category = category.replace("[]", "")
            else:
                self.category = category
            self.indices = indices

        def printInfo(self):
            return self.name + "  " + self.value

    class CoordSystemDefinition:
        def __init__(self, csNumber=None, definitions=[], forward=None, inverse=None):
            self.csNumber = csNumber  # coordinate system number
            self.motors = []  # list of motors assigned to CS
            self.motor = {}  # dict definition of each motor in terms of the axes
            # List of strings containing axes defs. I don't think these need
            # to be sorted.
            self.axisDefs = definitions
            self.parseAxesDefinitions(self.axisDefs)
            self.forward = forward  # instance of _KinematicTransform. Not implemented
            self.inverse = inverse  # instance of _KinematicTransform. Not implemented

        def parseAxesDefinitions(self, definitions):
            for definition in definitions:
                tokens = PPMACLexer(definition)
                motorNum = tokens.pop(3)[1]
                self.motors.append(motorNum)
                tokens.pop()
                self.motor[motorNum] = tokens.getTokensAsString()

        def printInfo(self):
            return "".join(
                [
                    f"&{self.csNumber}#{key}->{value}\n"
                    for key, value in self.motor.items()
                ]
            )

    def Program(self, name, size, offset, type, listing):
        return self._Program(self, name, size, offset, type, listing)

    class _Program:
        def __init__(self, ppmac, name, size, offset, type, listing):
            self.name = name
            self.size = size
            self.offset = offset
            self.type = type
            self.listing = listing
            self.dataStructureNames = set(ppmac.dataStructures.keys())
            # Currently dataStructures are not read from the repository
            # self.lexer = PPMACLexer('\n'.join(self.listing), self.dataStructureNames)
            # self.tokens = self.lexer.tokens

        def printInfo(self):
            s = (
                f"{self.type}, {self.name}, size {self.size}, offset {self.offset}\n"
                + "".join(self.listing)
            )
            return s

    def KinematicTransform(self, name, size, offset, type, coordSystem, listing):
        return self._KinematicTransform(
            self, name, size, offset, type, coordSystem, listing
        )

    class _KinematicTransform(_Program):
        def __init__(self, ppmac, name, size, offset, type, coordSystem, listing):
            super().__init__(ppmac, name, size, offset, type, listing)
            self.coodSystem = coordSystem

        def printInfo(self):
            return (
                f"{self.type}, {self.name}, size {self.size}, offset {self.offset}, cs"
                f" {self.coodSystem}\n"
            )

    def __init__(self, name="unknown"):
        # Source: hardware or respository
        self.source = name
        # Dictionary mapping DS names to dataStructure objects
        self.dataStructures = {}
        # Dictionary of active elements
        self.activeElements = {}
        # Dictionary of programs
        self.motionPrograms = {}
        # Dictionary of sub-programs
        self.subPrograms = {}
        # Dictionary of plc-programs
        self.plcPrograms = {}
        # Dictionary of programs
        self.forwardPrograms = {}
        # Dictionary of programs
        self.inversePrograms = {}
        # Dictionary of coord system definitions
        self.coordSystemDefs = {}
        # 16384 total I variables, values range from I0 to I16383
        # 16384 total P variables, values range from P0 to P65535
        # 16384 total M variables, values range from M0 to M16383
        # 8192 Q variables per coordinate system, values range from Q0 to Q8191
        # 8192 L variables per communications thread, coordinate system, or PLC;
        # values range from L0 to L8191
        # Power PMAC supports up to 32 PLC programs
        # Power PMAC supports up to 256 motors
        # BufIo[i] can have i = 0,...,63 (software reference manual p.541-542)
        # Power PMAC supports up to 256 cam tables
        # Power PMAC supports up to 256 compensation tables
        # Power PMAC supports up to 128 coordinate systems
        # Power PMAC supports up to 9 ECAT networks
        # Power PMAC supports up to 768 encoder conversion tables (software
        # reference manual p.206)
        # Gate1[i] can have i = 4,...,19 (software reference manual p.237)
        # Gate2[i] has an unknown range of indices i
        # self.numberOfGate2ICs = ???
        # Gate3[i] can have i = 0,..,15 (software reference manual p.289)
        # GateIo[i] can have i = 0,..,15 (software reference manual p.360),
        # although i > 15 does not return an error


class PPMACanalyse:
    def __init__(self, ppmacArgs):
        self.resultsDir = "ppmacAnalyse"
        self.verbosity = "info"
        self.ipAddress = "192.168.56.10"
        self.port = 1025
        self.operationType = "all"
        self.operationTypes = ["all", "active", "project"]
        self.backupDir = None
        self.username = None
        self.password = None
        if ppmacArgs.interface is not None:
            if not isValidNetworkInterface(ppmacArgs.interface[0]):
                raise IOError(
                    f"Interface {ppmacArgs.interface} not a valid network interface"
                    " <ipaddress>:<port>"
                )
            self.ipAddress = ppmacArgs.interface[0].split(":")[0]
            self.port = ppmacArgs.interface[0].split(":")[1]
        # Configure logger
        if ppmacArgs.resultsdir is not None:
            self.resultsDir = ppmacArgs.resultsdir[0]
        os.makedirs(self.resultsDir, exist_ok=True)
        logfile = self.resultsDir + "/" + "ppmacanalyse.log"
        logging.basicConfig(filename=logfile, level=logging.INFO)
        if ppmacArgs.username is not None:
            self.username = ppmacArgs.username
        if ppmacArgs.password is not None:
            self.password = ppmacArgs.password
        # Perform the backup, compare, recover or download
        if ppmacArgs.backup is not None:
            self.processBackupOptions(ppmacArgs)
            self.backup(self.operationType)
        if ppmacArgs.compare is not None:
            self.processCompareOptions(ppmacArgs)
            self.compare(self.operationType)
        if ppmacArgs.recover is not None:
            self.processRecoverOptions(ppmacArgs)
            self.recover()
        if ppmacArgs.download is not None:
            self.processDownloadOptions(ppmacArgs)
            self.download()

    def processCompareOptions(self, ppmacArgs):
        if len(ppmacArgs.compare) < 3:
            raise IOError(
                "Insufficient number of arguments, please specify compare option and"
                " TWO sources for comparison."
            )
        if ppmacArgs.compare[0] not in self.operationTypes:
            raise IOError(
                f"Unrecognised backup option {ppmacArgs.compare[0]}, "
                'should be "all","active" or "project".'
            )
        self.operationType = ppmacArgs.compare[0]
        self.compareSourceA = ppmacArgs.compare[1]
        self.compareSourceB = ppmacArgs.compare[2]
        if not isValidNetworkInterface(self.compareSourceA) and not os.path.isdir(
            self.compareSourceA
        ):
            raise IOError(
                f"{self.compareSourceA} not an existing directory or valid network"
                " interface <ipaddress>:<port>"
            )
        if not isValidNetworkInterface(self.compareSourceB) and not os.path.isdir(
            self.compareSourceB
        ):
            raise IOError(
                f"{self.compareSourceB} not an existing directory or valid network"
                " interface <ipaddress>:<port>"
            )
        self.compareDir = self.resultsDir
        os.makedirs(self.compareDir, exist_ok=True)
        self.ignoreFile = "ignore/ignore"
        if len(ppmacArgs.compare) > 3:
            self.ignoreFile = ppmacArgs.compare[3]
        if not fileExists(self.ignoreFile):
            raise IOError(f"Ignore file {self.ignoreFile} not found.")

    def compare(self, type="all"):
        ppmacA = PowerPMAC()
        ppmacB = PowerPMAC()
        if isValidNetworkInterface(self.compareSourceA):
            sshClient.hostname = self.compareSourceA.strip().split(":")[0]
            sshClient.port = self.compareSourceA.strip().split(":")[1]
            # Check that we can connect
            self.checkConnection(False)
            # sshClient.connect()
            if type == "all" or type == "active":
                hardwareWriteRead = PPMACHardwareWriteRead(
                    ppmacA, f"{self.compareDir}/tmp/databaseA"
                )
                hardwareWriteRead.readAndStoreActiveState(self.ignoreFile)
            if type == "all" or type == "project":
                projectASaved = PPMACProject(
                    "hardware",
                    "/opt/ppmac/usrflash/*",
                    f"{self.compareDir}/tmp/projectA/saved",
                )
                projectAActive = PPMACProject(
                    "hardware",
                    "/var/ftp/usrflash/*",
                    f"{self.compareDir}/tmp/projectA/active",
                )
            sshClient.disconnect()
        else:
            if type == "all" or type == "active":
                repositoryWriteRead = PPMACRepositoryWriteRead(
                    ppmacA, self.compareSourceA
                )
                repositoryWriteRead.readAndStoreActiveElements()
                repositoryWriteRead.readAndStoreBufferedPrograms()
                repositoryWriteRead.readAndStoreCSAxesDefinitions()
            if type == "all" or type == "project":
                projectASaved = PPMACProject(
                    "repository", f"{self.compareSourceA}/project/saved"
                )
                projectAActive = PPMACProject(
                    "repository", f"{self.compareSourceA}/project/active"
                )
        if isValidNetworkInterface(self.compareSourceB):
            sshClient.hostname = self.compareSourceB.strip().split(":")[0]
            sshClient.port = self.compareSourceB.strip().split(":")[1]
            # Check that we can connect
            self.checkConnection(False)
            # sshClient.connect()
            if type == "all" or type == "active":
                hardwareWriteRead = PPMACHardwareWriteRead(
                    ppmacB, f"{self.compareDir}/tmp/databaseB"
                )
                hardwareWriteRead.readAndStoreActiveState(self.ignoreFile)
            if type == "all" or type == "project":
                projectBSaved = PPMACProject(
                    "hardware",
                    "/opt/ppmac/usrflash/*",
                    f"{self.compareDir}/tmp/projectB/saved",
                )
                projectBActive = PPMACProject(
                    "hardware",
                    "/var/ftp/usrflash/*",
                    f"{self.compareDir}/tmp/projectB/active",
                )
            sshClient.disconnect()
        else:
            if type == "all" or type == "active":
                repositoryWriteRead = PPMACRepositoryWriteRead(
                    ppmacB, self.compareSourceB
                )
                repositoryWriteRead.readAndStoreActiveElements()
                repositoryWriteRead.readAndStoreBufferedPrograms()
                repositoryWriteRead.readAndStoreCSAxesDefinitions()
            if type == "all" or type == "project":
                projectBSaved = PPMACProject(
                    "repository", f"{self.compareSourceB}/project/saved"
                )
                projectBActive = PPMACProject(
                    "repository", f"{self.compareSourceB}/project/active"
                )
        # Run comparison
        ppmacComparison = PPMACCompare(ppmacA, ppmacB, self.compareDir)
        if type == "all" or type == "active":
            ppmacComparison.compareActiveElements()
            ppmacComparison.comparePrograms()
            ppmacComparison.compareCoordSystemAxesDefinitions()
        if type == "all" or type == "project":
            savedProjComparison = ProjectCompare(projectASaved, projectBSaved)
            savedProjComparison.compareProjectFiles(f"{self.compareDir}/project/saved")
            activeProjComparison = ProjectCompare(projectAActive, projectBActive)
            activeProjComparison.compareProjectFiles(
                f"{self.compareDir}/project/active"
            )

    def processBackupOptions(self, ppmacArgs):
        self.name = "hardware"
        if ppmacArgs.name is not None:
            self.name = ppmacArgs.name[0]
        if len(ppmacArgs.backup) > 0:
            if ppmacArgs.backup[0] not in self.operationTypes:
                raise IOError(
                    f"Unrecognised backup option {ppmacArgs.backup[0]}, "
                    'should be "all","active" or "project".'
                )
            self.operationType = ppmacArgs.backup[0]
        self.backupDir = self.resultsDir
        os.makedirs(self.backupDir, exist_ok=True)
        self.ignoreFile = "ignore/ignore"
        if len(ppmacArgs.backup) > 1:
            self.ignoreFile = ppmacArgs.backup[1]
        if not fileExists(self.ignoreFile):
            raise IOError(f"Ignore file {self.ignoreFile} not found.")

    @connectDisconnect
    def backup(self, type="all"):
        ppmacA = PowerPMAC(self.name)
        # Check that we can connect
        self.checkConnection(False)
        if type == "all" or type == "active":
            # read current state of ppmac and store in ppmacA object
            hardwareWriteRead = PPMACHardwareWriteRead(ppmacA, f"{self.backupDir}/tmp")
            hardwareWriteRead.readAndStoreActiveState(self.ignoreFile)
            # write current state of ppmacA object to repository
            activeDir = self.backupDir
            repositoryWriteRead = PPMACRepositoryWriteRead(ppmacA, activeDir)
            repositoryWriteRead.writeActiveState()
        if type == "all" or type == "project":
            savedProjectDir = f"{self.backupDir}/project/saved"
            os.makedirs(savedProjectDir, exist_ok=True)
            scpFromPowerPMACtoLocal(
                "/opt/ppmac/usrflash/*", savedProjectDir, recursive=True
            )
            activeProjectDir = f"{self.backupDir}/project/active"
            os.makedirs(activeProjectDir, exist_ok=True)
            scpFromPowerPMACtoLocal(
                "/var/ftp/usrflash/*", activeProjectDir, recursive=True
            )

    def checkConnection(self, disconnectAfter):
        if self.username is not None and self.password is not None:
            logging.info(self.username)
            connect_status = sshClient.connect(
                username=self.username[0], password=self.password[0]
            )
        else:
            connect_status = sshClient.connect()
        if connect_status is None:
            # All OK
            return
        if "Invalid username or password" in connect_status:
            raise IOError("ERROR: Invalid username and/or password.")
        if "Cannot connect" in connect_status:
            logging.error(
                "Cannot establish connection to Power PMAC at "
                + str(self.ipAddress)
                + ":"
                + str(self.port)
            )
            raise IOError(
                f"ERROR: Cannot establish connection to Power PMAC at "
                f" {self.ipAddress}:{self.port}"
            )
        if disconnectAfter:
            sshClient.disconnect()

    def processRecoverOptions(self, ppmacArgs):
        # Specify directory containing backup
        if len(ppmacArgs.recover) < 1:
            raise IOError("Unspecified directory to use for recovery.")
        self.backupDir = ppmacArgs.recover[0]
        if not os.path.isdir(self.backupDir):
            raise IOError(f"Repository directory {self.backupDir} does not exist.")

    @connectDisconnect
    def recover(self):
        # Check that we can connect
        self.checkConnection(False)
        recoverScript = "recover.sh"
        with open(recoverScript, "w+") as writeFile:
            writeFile.write(recoveryCmds)
        scpFromLocalToPowerPMAC(recoverScript, "/tmp/")
        os.system("rm recover.sh")
        executeRemoteShellCommand(f"chmod 777 /tmp/{recoverScript}")
        executeRemoteShellCommand("mkdir -p /tmp/recover")
        scpFromLocalToPowerPMAC(
            f"{self.backupDir}/Project", "/tmp/recover/", recursive=True
        )
        scpFromLocalToPowerPMAC(
            f"{self.backupDir}/Database", "/tmp/recover/", recursive=True
        )
        scpFromLocalToPowerPMAC(
            f"{self.backupDir}/Temp", "/tmp/recover/", recursive=True
        )
        executeRemoteShellCommand("/tmp/recover.sh")
        scpFromPowerPMACtoLocal(
            "/tmp/recover.log", f"{self.resultsDir}/", recursive=False
        )

    def processDownloadOptions(self, ppmacArgs):
        # Specify directory containing backup
        if len(ppmacArgs.download) < 1:
            raise IOError("Unspecified directory to use for recovery.")
        self.backupDir = ppmacArgs.download[0]
        if not os.path.isdir(self.backupDir):
            raise IOError(f"Repository directory {self.backupDir} does not exist.")

    @connectDisconnect
    def download(self):
        # Check that we can connect
        self.checkConnection(False)
        # Copy usrflash files into ppmac
        executeRemoteShellCommand("rm -rf /var/ftp/usrflash/Project/*")
        for file in os.listdir(f"{self.backupDir}/Project"):
            scpFromLocalToPowerPMAC(
                f"{self.backupDir}/Project/{file}",
                "/var/ftp/usrflash/Project",
                recursive=True,
            )
        # Make directory that projpp will log to
        executeRemoteShellCommand("mkdir -p /var/ftp/usrflash/Project/Log")
        # Looks like everything in project dir needs to be rwx
        executeRemoteShellCommand("chmod 777 -R -f /var/ftp/usrflash/Project")
        # Finally execute a projpp to parse and load new project
        executeRemoteShellCommand("projpp -l")


sshClient = dls_pmacremote.PPmacSshInterface()


def main():
    """Main entry point of the script."""
    if len(sys.argv) < 2:
        print(
            " -> Run dls-ppmac-analyse --help for details on how to"
            + " use the Power-Pmac analyse tool"
        )
        return
    ppmacArgs = parseArgs()
    if ppmacArgs.gui:
        dls_powerpmacanalyse.ppmacanalyse_control.main()
    else:
        PPMACanalyse(ppmacArgs)


if __name__ == "__main__":
    main()



import signal
import subprocess
import sys
import time
from pathlib import Path

from PyQt5 import QtWidgets
from PyQt5.QtGui import QColor, QTextCursor
from PyQt5.QtWidgets import QApplication, QFileDialog, QMainWindow

from dls_powerpmacanalyse.htmlDisp import ComparisonViewer
from dls_powerpmacanalyse.login import Loginform
from dls_powerpmacanalyse.ui_formAnalyseControl import Ui_ControlForm


class Controlform(QtWidgets.QMainWindow, Ui_ControlForm):
    def __init__(self, parent=None):
        # signal.signal(2, self.signalHandler)
        # setup signals

        QMainWindow.__init__(self, parent)
        self.setupUi(self)
        self.login = Loginform(self)
        self.compviewer = ComparisonViewer(self)

        # Mode - set by the tab index of the GUI
        # 0 = backup
        # 1 = compare
        # 2 = download/recover
        self.mode = 0

        # Text colors
        self.blackColor = QColor(0, 0, 0)
        self.blueColor = QColor(0, 0, 255)
        self.redColor = QColor(255, 0, 0)

        # IP/Port
        # 0 = backup
        # 2 = download/recover
        server = "192.168.56.10"
        port = "22"
        self.lineServer0.setText(server)
        self.linePort0.setText(port)
        self.lineServer2.setText(server)
        self.linePort2.setText(port)

        # Back-up options
        self.backupCompareOption = "all"
        self.rdbAll0.setChecked(True)
        self.rdbAll1.setChecked(True)

        # Sources for compare
        source1 = "192.168.56.10:22"
        source2 = "./"
        self.lineSource1.setText(source1)
        self.lineSource2.setText(source2)

        # Ignore file location
        # 0 = backup
        # 1 = compare
        ignoreFile = "/dls_sw/work/motion/PPMAC_TEST/ignore"
        self.lineIgnoreFile0.setText(ignoreFile)
        self.lineIgnoreFile1.setText(ignoreFile)

        # Results file location
        # 0 = backup
        # 1 = compare
        # 2 = download/recover
        outputLocation = "./ppmacAnalyse"
        self.lineOutputDir0.setText(outputLocation)
        self.lineOutputDir1.setText(outputLocation)
        self.lineOutputDir2.setText(outputLocation)

        # Backup file location for download/recover
        self.lineBackupDir.setText("./")

        # Cancel buttons
        # 0 = backup
        # 1 = compare
        # 2 = download/recover
        self.pushCancel0.setEnabled(False)
        self.pushCancel1.setEnabled(False)
        self.pushCancel2.setEnabled(False)
        self.cancelBackup = False
        self.cancelCompare = False
        self.cancelDR = False

    def runBackup(self):
        # Tab index 0
        server_name = self.lineServer0.text()
        server_port = self.linePort0.text()
        ignore_file = self.lineIgnoreFile0.text()
        output_dir = self.lineOutputDir0.text()
        backup_option = "all"
        if self.rdbProject0.isChecked():
            backup_option = "project"
        elif self.rdbActive0.isChecked():
            backup_option = "active"

        cmd0 = [
            "dls-powerpmac-analyse",
            "--interface",
            str(server_name) + ":" + str(server_port),
            "--backup",
            backup_option,
            str(ignore_file),
            "--resultsdir",
            str(output_dir),
        ]

        self.runPPmacAnalyseCmd(cmd0, 0, "Backup")

    def runCompare(self):
        # Tab index 1
        source1 = self.lineSource1.text()
        source2 = self.lineSource2.text()
        ignore_file = self.lineIgnoreFile1.text()
        output_dir = self.lineOutputDir1.text()
        compare_option = "all"
        if self.rdbProject1.isChecked():
            compare_option = "project"
        elif self.rdbActive1.isChecked():
            compare_option = "active"

        cmd0 = [
            "dls-powerpmac-analyse",
            "--compare",
            compare_option,
            source1,
            source2,
            str(ignore_file),
            "--resultsdir",
            str(output_dir),
        ]

        self.runPPmacAnalyseCmd(cmd0, 1, "Compare")

    def runDownload(self):
        # Tab index 2
        server_name = self.lineServer2.text()
        server_port = self.linePort2.text()
        backup_dir = self.lineBackupDir.text()
        output_dir = self.lineOutputDir2.text()

        backup_dir += "/project/active"

        cmd0 = [
            "dls-powerpmac-analyse",
            "--interface",
            str(server_name) + ":" + str(server_port),
            "--download",
            str(backup_dir),
            "--resultsdir",
            str(output_dir),
        ]

        self.runPPmacAnalyseCmd(cmd0, 2, "Download")

    def runRecover(self):
        # Tab index 2
        server_name = self.lineServer2.text()
        server_port = self.linePort2.text()
        backup_dir = self.lineBackupDir.text()
        output_dir = self.lineOutputDir2.text()

        backup_dir += "/project/saved"

        cmd0 = [
            "dls-powerpmac-analyse",
            "--interface",
            str(server_name) + ":" + str(server_port),
            "--recover",
            str(backup_dir),
            "--resultsdir",
            str(output_dir),
        ]

        self.runPPmacAnalyseCmd(cmd0, 2, "Recover")

    def runPPmacAnalyseCmd(self, cmd0, guiTab, optionStr):
        self.setInternalCancelled(guiTab, False)

        finished = False
        count = 0
        cmdRerun = cmd0
        # Run until we signal that it should finish
        while not finished:
            if count == 0:
                cmd = cmd0
            else:
                cmd = cmdRerun

            self.enableCancelButton(guiTab, True)
            self.addTextLog("Running cmd: '" + str(" ".join(cmd)) + "'")
            start = time.time()

            process = subprocess.Popen(
                cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            self.addTextProgress("Working .")

            logInterval = time.time()

            while process.poll() is None:
                # Only log every second
                if time.time() - logInterval > 5:
                    self.addTextProgress(".")
                    logInterval = time.time()
                QApplication.processEvents()
                if self.wasCancelled(guiTab):
                    self.addTextError("\nCancelling")
                    process.kill()
                time.sleep(0.1)

            self.enableCancelButton(guiTab, False)
            success = True
            stderrStr = ""
            for line in process.stderr:
                unicode_text = str(line, "utf-8")
                if unicode_text != "":
                    success = False
                    # Just save the last line with the root error
                    stderrStr = unicode_text

            if "Invalid username" in stderrStr:
                self.addTextError("\nBackup failed with errors: \n" + stderrStr)
                is_clickedOK = self.login.exec()
                if is_clickedOK:
                    new_cmds = [
                        "--username",
                        str(self.login.username),
                        "--password",
                        str(self.login.password),
                    ]
                    cmdRerun = cmd0.copy()
                    for new_cmd in new_cmds:
                        cmdRerun.append(new_cmd)
                else:
                    finished = True
            else:
                finished = True

            count += 1

        # stdout, stderr = process.communicate()
        if self.wasCancelled(guiTab):
            self.addTextLog(optionStr + " cancelled...")
            self.setInternalCancelled(guiTab, False)
        elif not success:
            self.addTextError("\n" + optionStr + " failed with errors: \n" + stderrStr)
        else:
            self.addTextLog(
                "\n"
                + optionStr
                + " completed in: "
                + str(time.time() - start)
                + " secs"
            )

    def enableCancelButton(self, guiTab, value):
        if guiTab == 0:
            self.pushCancel0.setEnabled(value)
        elif guiTab == 1:
            self.pushCancel1.setEnabled(value)
        elif guiTab == 2:
            self.pushCancel2.setEnabled(value)

    def wasCancelled(self, guiTab):
        if guiTab == 0:
            return self.cancelBackup
        elif guiTab == 1:
            return self.cancelCompare
        elif guiTab == 2:
            return self.cancelDR

    def setInternalCancelled(self, guiTab, value):
        if guiTab == 0:
            self.cancelBackup = value
        elif guiTab == 1:
            self.cancelCompare = value
        elif guiTab == 2:
            self.cancelDR = value

    def cancelBackup(self):
        self.cancelBackup = True

    def cancelCompare(self):
        self.cancelCompare = True

    def cancelDR(self):
        self.cancelDR = True

    def ignoreFileBrowser(self):
        filename, _filter = QFileDialog.getOpenFileName()
        if filename != "":
            if self.mode == 0:
                self.lineIgnoreFile.setText(filename)
            elif self.mode == 1:
                self.lineIgnoreFile2.setText(filename)

    def outputDirBrowser(self):
        directory = QFileDialog.getExistingDirectory()
        if directory != "":
            if self.mode == 0:
                self.lineOutputDir0.setText(directory)
            elif self.mode == 1:
                self.lineOutputDir1.setText(directory)
            elif self.mode == 2:
                self.lineOutputDir2.setText(directory)

    def backupDirBrowser(self):
        directory = QFileDialog.getExistingDirectory()
        if directory != "":
            self.lineBackupDir.setText(directory)

    def source1LocationBrowser(self):
        directory = QFileDialog.getExistingDirectory()
        if directory != "":
            self.lineSource1.setText(directory)

    def source2LocationBrowser(self):
        directory = QFileDialog.getExistingDirectory()
        if directory != "":
            self.lineSource2.setText(directory)

    def addTextLog(self, text):
        self.addTxtToOuput(text, True, self.blackColor)

    def addTextProgress(self, text):
        self.addTxtToOuput(text, False, self.blueColor)

    def addTextError(self, text):
        self.addTxtToOuput(text, True, self.redColor)

    def addTxtToOuput(self, text, insertNewLine, color):
        self.textOutput.setTextColor(color)
        if insertNewLine:
            self.textOutput.insertPlainText(text + "\n")
        else:
            self.textOutput.insertPlainText(text)
        self.textOutput.moveCursor(QTextCursor.End)

    def setMode(self, tabId):
        self.mode = tabId

    def openViewer(self):
        output_dir = self.lineOutputDir1.text()
        fileList = [_.as_posix() for _ in Path(output_dir).glob("**/*html")]
        self.compviewer.setup(output_dir, fileList)
        self.compviewer.show()


def main():
    app = QApplication(sys.argv)
    app.lastWindowClosed.connect(app.quit)
    win = Controlform()
    win.show()
    # catch CTRL-C
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    app.exec_()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file
# 'dls_powerpmacanalyse/formAnalyseControl.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtWidgets


class Ui_ControlForm(object):
    def setupUi(self, ControlForm):
        ControlForm.setObjectName("ControlForm")
        ControlForm.resize(1038, 598)
        self.centralwidget = QtWidgets.QWidget(ControlForm)
        self.centralwidget.setObjectName("centralwidget")
        self.textOutput = QtWidgets.QTextEdit(self.centralwidget)
        self.textOutput.setGeometry(QtCore.QRect(20, 270, 991, 281))
        self.textOutput.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.textOutput.setReadOnly(True)
        self.textOutput.setObjectName("textOutput")
        self.tabWidget = QtWidgets.QTabWidget(self.centralwidget)
        self.tabWidget.setGeometry(QtCore.QRect(20, 20, 991, 241))
        self.tabWidget.setObjectName("tabWidget")
        self.Backup = QtWidgets.QWidget()
        self.Backup.setObjectName("Backup")
        self.groupBox_4 = QtWidgets.QGroupBox(self.Backup)
        self.groupBox_4.setGeometry(QtCore.QRect(10, 10, 971, 191))
        self.groupBox_4.setObjectName("groupBox_4")
        self.groupBox = QtWidgets.QGroupBox(self.groupBox_4)
        self.groupBox.setGeometry(QtCore.QRect(10, 30, 241, 151))
        self.groupBox.setObjectName("groupBox")
        self.layoutWidget = QtWidgets.QWidget(self.groupBox)
        self.layoutWidget.setGeometry(QtCore.QRect(10, 30, 220, 80))
        self.layoutWidget.setObjectName("layoutWidget")
        self.gridLayout = QtWidgets.QGridLayout(self.layoutWidget)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.textLabel1 = QtWidgets.QLabel(self.layoutWidget)
        self.textLabel1.setMinimumSize(QtCore.QSize(75, 27))
        self.textLabel1.setMaximumSize(QtCore.QSize(75, 25))
        self.textLabel1.setWordWrap(False)
        self.textLabel1.setObjectName("textLabel1")
        self.gridLayout.addWidget(self.textLabel1, 0, 0, 1, 1)
        self.lineServer0 = QtWidgets.QLineEdit(self.layoutWidget)
        self.lineServer0.setObjectName("lineServer0")
        self.gridLayout.addWidget(self.lineServer0, 0, 1, 1, 1)
        self.textLabel2 = QtWidgets.QLabel(self.layoutWidget)
        self.textLabel2.setMinimumSize(QtCore.QSize(75, 27))
        self.textLabel2.setMaximumSize(QtCore.QSize(75, 27))
        self.textLabel2.setWordWrap(False)
        self.textLabel2.setObjectName("textLabel2")
        self.gridLayout.addWidget(self.textLabel2, 1, 0, 1, 1)
        self.linePort0 = QtWidgets.QLineEdit(self.layoutWidget)
        self.linePort0.setObjectName("linePort0")
        self.gridLayout.addWidget(self.linePort0, 1, 1, 1, 1)
        self.groupBox_2 = QtWidgets.QGroupBox(self.groupBox_4)
        self.groupBox_2.setGeometry(QtCore.QRect(260, 30, 541, 151))
        self.groupBox_2.setObjectName("groupBox_2")
        self.layoutWidget1 = QtWidgets.QWidget(self.groupBox_2)
        self.layoutWidget1.setGeometry(QtCore.QRect(10, 31, 77, 92))
        self.layoutWidget1.setObjectName("layoutWidget1")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.layoutWidget1)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.rdbActive0 = QtWidgets.QRadioButton(self.layoutWidget1)
        self.rdbActive0.setObjectName("rdbActive0")
        self.verticalLayout.addWidget(self.rdbActive0)
        self.rdbProject0 = QtWidgets.QRadioButton(self.layoutWidget1)
        self.rdbProject0.setObjectName("rdbProject0")
        self.verticalLayout.addWidget(self.rdbProject0)
        self.rdbAll0 = QtWidgets.QRadioButton(self.layoutWidget1)
        self.rdbAll0.setObjectName("rdbAll0")
        self.verticalLayout.addWidget(self.rdbAll0)
        self.layoutWidget2 = QtWidgets.QWidget(self.groupBox_2)
        self.layoutWidget2.setGeometry(QtCore.QRect(100, 30, 431, 80))
        self.layoutWidget2.setObjectName("layoutWidget2")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.layoutWidget2)
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.label = QtWidgets.QLabel(self.layoutWidget2)
        self.label.setObjectName("label")
        self.gridLayout_2.addWidget(self.label, 0, 0, 1, 1)
        self.lineIgnoreFile0 = QtWidgets.QLineEdit(self.layoutWidget2)
        self.lineIgnoreFile0.setObjectName("lineIgnoreFile0")
        self.gridLayout_2.addWidget(self.lineIgnoreFile0, 0, 1, 1, 2)
        self.pushIgnoreLocate = QtWidgets.QPushButton(self.layoutWidget2)
        self.pushIgnoreLocate.setObjectName("pushIgnoreLocate")
        self.gridLayout_2.addWidget(self.pushIgnoreLocate, 0, 3, 1, 1)
        self.label_2 = QtWidgets.QLabel(self.layoutWidget2)
        self.label_2.setObjectName("label_2")
        self.gridLayout_2.addWidget(self.label_2, 1, 0, 1, 2)
        self.lineOutputDir0 = QtWidgets.QLineEdit(self.layoutWidget2)
        self.lineOutputDir0.setObjectName("lineOutputDir0")
        self.gridLayout_2.addWidget(self.lineOutputDir0, 1, 2, 1, 1)
        self.pushDirLocate = QtWidgets.QPushButton(self.layoutWidget2)
        self.pushDirLocate.setObjectName("pushDirLocate")
        self.gridLayout_2.addWidget(self.pushDirLocate, 1, 3, 1, 1)
        self.groupBox_3 = QtWidgets.QGroupBox(self.groupBox_4)
        self.groupBox_3.setGeometry(QtCore.QRect(810, 30, 141, 151))
        self.groupBox_3.setObjectName("groupBox_3")
        self.pushBackup = QtWidgets.QPushButton(self.groupBox_3)
        self.pushBackup.setGeometry(QtCore.QRect(10, 30, 92, 36))
        self.pushBackup.setObjectName("pushBackup")
        self.pushCancel0 = QtWidgets.QPushButton(self.groupBox_3)
        self.pushCancel0.setGeometry(QtCore.QRect(10, 70, 92, 36))
        self.pushCancel0.setObjectName("pushCancel0")
        self.tabWidget.addTab(self.Backup, "")
        self.tab_2 = QtWidgets.QWidget()
        self.tab_2.setObjectName("tab_2")
        self.groupBox_5 = QtWidgets.QGroupBox(self.tab_2)
        self.groupBox_5.setGeometry(QtCore.QRect(10, 10, 971, 191))
        self.groupBox_5.setObjectName("groupBox_5")
        self.groupBox_6 = QtWidgets.QGroupBox(self.groupBox_5)
        self.groupBox_6.setGeometry(QtCore.QRect(10, 30, 351, 151))
        self.groupBox_6.setObjectName("groupBox_6")
        self.layoutWidget3 = QtWidgets.QWidget(self.groupBox_6)
        self.layoutWidget3.setGeometry(QtCore.QRect(11, 28, 331, 80))
        self.layoutWidget3.setObjectName("layoutWidget3")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.layoutWidget3)
        self.gridLayout_3.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.textLabel1_2 = QtWidgets.QLabel(self.layoutWidget3)
        self.textLabel1_2.setMinimumSize(QtCore.QSize(75, 27))
        self.textLabel1_2.setMaximumSize(QtCore.QSize(75, 25))
        self.textLabel1_2.setWordWrap(False)
        self.textLabel1_2.setObjectName("textLabel1_2")
        self.gridLayout_3.addWidget(self.textLabel1_2, 0, 0, 1, 1)
        self.lineSource1 = QtWidgets.QLineEdit(self.layoutWidget3)
        self.lineSource1.setObjectName("lineSource1")
        self.gridLayout_3.addWidget(self.lineSource1, 0, 1, 1, 1)
        self.pushDirLocate_4 = QtWidgets.QPushButton(self.layoutWidget3)
        self.pushDirLocate_4.setObjectName("pushDirLocate_4")
        self.gridLayout_3.addWidget(self.pushDirLocate_4, 0, 2, 1, 1)
        self.textLabel1_3 = QtWidgets.QLabel(self.layoutWidget3)
        self.textLabel1_3.setMinimumSize(QtCore.QSize(75, 27))
        self.textLabel1_3.setMaximumSize(QtCore.QSize(75, 25))
        self.textLabel1_3.setWordWrap(False)
        self.textLabel1_3.setObjectName("textLabel1_3")
        self.gridLayout_3.addWidget(self.textLabel1_3, 1, 0, 1, 1)
        self.lineSource2 = QtWidgets.QLineEdit(self.layoutWidget3)
        self.lineSource2.setObjectName("lineSource2")
        self.gridLayout_3.addWidget(self.lineSource2, 1, 1, 1, 1)
        self.pushDirLocate_3 = QtWidgets.QPushButton(self.layoutWidget3)
        self.pushDirLocate_3.setObjectName("pushDirLocate_3")
        self.gridLayout_3.addWidget(self.pushDirLocate_3, 1, 2, 1, 1)
        self.groupBox_7 = QtWidgets.QGroupBox(self.groupBox_5)
        self.groupBox_7.setGeometry(QtCore.QRect(370, 30, 471, 151))
        self.groupBox_7.setObjectName("groupBox_7")
        self.layoutWidget_3 = QtWidgets.QWidget(self.groupBox_7)
        self.layoutWidget_3.setGeometry(QtCore.QRect(97, 30, 361, 80))
        self.layoutWidget_3.setObjectName("layoutWidget_3")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.layoutWidget_3)
        self.gridLayout_4.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.label_3 = QtWidgets.QLabel(self.layoutWidget_3)
        self.label_3.setObjectName("label_3")
        self.gridLayout_4.addWidget(self.label_3, 0, 0, 1, 1)
        self.lineIgnoreFile1 = QtWidgets.QLineEdit(self.layoutWidget_3)
        self.lineIgnoreFile1.setObjectName("lineIgnoreFile1")
        self.gridLayout_4.addWidget(self.lineIgnoreFile1, 0, 1, 1, 2)
        self.pushIgnoreLocate_2 = QtWidgets.QPushButton(self.layoutWidget_3)
        self.pushIgnoreLocate_2.setObjectName("pushIgnoreLocate_2")
        self.gridLayout_4.addWidget(self.pushIgnoreLocate_2, 0, 3, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.layoutWidget_3)
        self.label_4.setObjectName("label_4")
        self.gridLayout_4.addWidget(self.label_4, 1, 0, 1, 2)
        self.lineOutputDir1 = QtWidgets.QLineEdit(self.layoutWidget_3)
        self.lineOutputDir1.setObjectName("lineOutputDir1")
        self.gridLayout_4.addWidget(self.lineOutputDir1, 1, 2, 1, 1)
        self.pushDirLocate_2 = QtWidgets.QPushButton(self.layoutWidget_3)
        self.pushDirLocate_2.setObjectName("pushDirLocate_2")
        self.gridLayout_4.addWidget(self.pushDirLocate_2, 1, 3, 1, 1)
        self.layoutWidget4 = QtWidgets.QWidget(self.groupBox_7)
        self.layoutWidget4.setGeometry(QtCore.QRect(8, 30, 77, 92))
        self.layoutWidget4.setObjectName("layoutWidget4")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.layoutWidget4)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.rdbActive1 = QtWidgets.QRadioButton(self.layoutWidget4)
        self.rdbActive1.setObjectName("rdbActive1")
        self.verticalLayout_2.addWidget(self.rdbActive1)
        self.rdbProject1 = QtWidgets.QRadioButton(self.layoutWidget4)
        self.rdbProject1.setObjectName("rdbProject1")
        self.verticalLayout_2.addWidget(self.rdbProject1)
        self.rdbAll1 = QtWidgets.QRadioButton(self.layoutWidget4)
        self.rdbAll1.setObjectName("rdbAll1")
        self.verticalLayout_2.addWidget(self.rdbAll1)
        self.pushOpenViewer = QtWidgets.QPushButton(self.groupBox_7)
        self.pushOpenViewer.setGeometry(QtCore.QRect(260, 110, 201, 30))
        self.pushOpenViewer.setObjectName("pushOpenViewer")
        self.groupBox_8 = QtWidgets.QGroupBox(self.groupBox_5)
        self.groupBox_8.setGeometry(QtCore.QRect(851, 30, 111, 141))
        self.groupBox_8.setObjectName("groupBox_8")
        self.pushCompare = QtWidgets.QPushButton(self.groupBox_8)
        self.pushCompare.setGeometry(QtCore.QRect(10, 30, 92, 36))
        self.pushCompare.setObjectName("pushCompare")
        self.pushCancel1 = QtWidgets.QPushButton(self.groupBox_8)
        self.pushCancel1.setGeometry(QtCore.QRect(10, 70, 92, 36))
        self.pushCancel1.setObjectName("pushCancel1")
        self.tabWidget.addTab(self.tab_2, "")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.groupBox_9 = QtWidgets.QGroupBox(self.tab)
        self.groupBox_9.setGeometry(QtCore.QRect(10, 10, 971, 191))
        self.groupBox_9.setObjectName("groupBox_9")
        self.groupBox_10 = QtWidgets.QGroupBox(self.groupBox_9)
        self.groupBox_10.setGeometry(QtCore.QRect(10, 30, 241, 141))
        self.groupBox_10.setObjectName("groupBox_10")
        self.layoutWidget_2 = QtWidgets.QWidget(self.groupBox_10)
        self.layoutWidget_2.setGeometry(QtCore.QRect(10, 30, 220, 80))
        self.layoutWidget_2.setObjectName("layoutWidget_2")
        self.gridLayout_5 = QtWidgets.QGridLayout(self.layoutWidget_2)
        self.gridLayout_5.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.textLabel1_4 = QtWidgets.QLabel(self.layoutWidget_2)
        self.textLabel1_4.setMinimumSize(QtCore.QSize(75, 27))
        self.textLabel1_4.setMaximumSize(QtCore.QSize(75, 25))
        self.textLabel1_4.setWordWrap(False)
        self.textLabel1_4.setObjectName("textLabel1_4")
        self.gridLayout_5.addWidget(self.textLabel1_4, 0, 0, 1, 1)
        self.lineServer2 = QtWidgets.QLineEdit(self.layoutWidget_2)
        self.lineServer2.setObjectName("lineServer2")
        self.gridLayout_5.addWidget(self.lineServer2, 0, 1, 1, 1)
        self.textLabel2_2 = QtWidgets.QLabel(self.layoutWidget_2)
        self.textLabel2_2.setMinimumSize(QtCore.QSize(75, 27))
        self.textLabel2_2.setMaximumSize(QtCore.QSize(75, 27))
        self.textLabel2_2.setWordWrap(False)
        self.textLabel2_2.setObjectName("textLabel2_2")
        self.gridLayout_5.addWidget(self.textLabel2_2, 1, 0, 1, 1)
        self.linePort2 = QtWidgets.QLineEdit(self.layoutWidget_2)
        self.linePort2.setObjectName("linePort2")
        self.gridLayout_5.addWidget(self.linePort2, 1, 1, 1, 1)
        self.groupBox_11 = QtWidgets.QGroupBox(self.groupBox_9)
        self.groupBox_11.setGeometry(QtCore.QRect(260, 30, 481, 141))
        self.groupBox_11.setObjectName("groupBox_11")
        self.layoutWidget_5 = QtWidgets.QWidget(self.groupBox_11)
        self.layoutWidget_5.setGeometry(QtCore.QRect(10, 30, 461, 80))
        self.layoutWidget_5.setObjectName("layoutWidget_5")
        self.gridLayout_6 = QtWidgets.QGridLayout(self.layoutWidget_5)
        self.gridLayout_6.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_6.setObjectName("gridLayout_6")
        self.label_5 = QtWidgets.QLabel(self.layoutWidget_5)
        self.label_5.setObjectName("label_5")
        self.gridLayout_6.addWidget(self.label_5, 0, 0, 1, 1)
        self.lineBackupDir = QtWidgets.QLineEdit(self.layoutWidget_5)
        self.lineBackupDir.setObjectName("lineBackupDir")
        self.gridLayout_6.addWidget(self.lineBackupDir, 0, 1, 1, 2)
        self.pushBackupDir = QtWidgets.QPushButton(self.layoutWidget_5)
        self.pushBackupDir.setObjectName("pushBackupDir")
        self.gridLayout_6.addWidget(self.pushBackupDir, 0, 3, 1, 1)
        self.label_6 = QtWidgets.QLabel(self.layoutWidget_5)
        self.label_6.setObjectName("label_6")
        self.gridLayout_6.addWidget(self.label_6, 1, 0, 1, 2)
        self.lineOutputDir2 = QtWidgets.QLineEdit(self.layoutWidget_5)
        self.lineOutputDir2.setObjectName("lineOutputDir2")
        self.gridLayout_6.addWidget(self.lineOutputDir2, 1, 2, 1, 1)
        self.pushOutputDir = QtWidgets.QPushButton(self.layoutWidget_5)
        self.pushOutputDir.setObjectName("pushOutputDir")
        self.gridLayout_6.addWidget(self.pushOutputDir, 1, 3, 1, 1)
        self.groupBox_12 = QtWidgets.QGroupBox(self.groupBox_9)
        self.groupBox_12.setGeometry(QtCore.QRect(750, 30, 211, 141))
        self.groupBox_12.setObjectName("groupBox_12")
        self.pushDownload = QtWidgets.QPushButton(self.groupBox_12)
        self.pushDownload.setGeometry(QtCore.QRect(10, 30, 92, 36))
        self.pushDownload.setObjectName("pushDownload")
        self.pushCancel2 = QtWidgets.QPushButton(self.groupBox_12)
        self.pushCancel2.setGeometry(QtCore.QRect(60, 70, 92, 36))
        self.pushCancel2.setObjectName("pushCancel2")
        self.pushRecover = QtWidgets.QPushButton(self.groupBox_12)
        self.pushRecover.setGeometry(QtCore.QRect(110, 30, 92, 36))
        self.pushRecover.setObjectName("pushRecover")
        self.tabWidget.addTab(self.tab, "")
        ControlForm.setCentralWidget(self.centralwidget)
        self.statusbar = QtWidgets.QStatusBar(ControlForm)
        self.statusbar.setObjectName("statusbar")
        ControlForm.setStatusBar(self.statusbar)

        self.retranslateUi(ControlForm)
        self.tabWidget.setCurrentIndex(0)
        self.pushBackup.clicked.connect(ControlForm.runBackup)
        self.pushIgnoreLocate.clicked.connect(ControlForm.ignoreFileBrowser)
        self.pushCompare.clicked.connect(ControlForm.runCompare)
        self.tabWidget.currentChanged["int"].connect(ControlForm.setMode)
        self.pushDirLocate_4.clicked.connect(ControlForm.source1LocationBrowser)
        self.pushDirLocate_3.clicked.connect(ControlForm.source2LocationBrowser)
        self.pushIgnoreLocate_2.clicked.connect(ControlForm.ignoreFileBrowser)
        self.pushCancel0.clicked.connect(ControlForm.cancelBackup)
        self.pushCancel1.clicked.connect(ControlForm.cancelCompare)
        self.pushDownload.clicked.connect(ControlForm.runDownload)
        self.pushRecover.clicked.connect(ControlForm.runRecover)
        self.pushCancel2.clicked.connect(ControlForm.cancelDR)
        self.pushDirLocate.clicked.connect(ControlForm.outputDirBrowser)
        self.pushDirLocate_2.clicked.connect(ControlForm.outputDirBrowser)
        self.pushBackupDir.clicked.connect(ControlForm.backupDirBrowser)
        self.pushOutputDir.clicked.connect(ControlForm.outputDirBrowser)
        self.pushOpenViewer.clicked.connect(ControlForm.openViewer)
        QtCore.QMetaObject.connectSlotsByName(ControlForm)

    def retranslateUi(self, ControlForm):
        _translate = QtCore.QCoreApplication.translate
        ControlForm.setWindowTitle(_translate("ControlForm", "Power-PMAC Analyse Tool"))
        self.groupBox_4.setTitle(
            _translate("ControlForm", "Power PMAC Backup Configuration")
        )
        self.groupBox.setTitle(_translate("ControlForm", "Connection to Power PMAC"))
        self.textLabel1.setText(_translate("ControlForm", "Server:"))
        self.lineServer0.setToolTip(
            _translate(
                "ControlForm", "Specify the IP address of the Power PMAC to backup"
            )
        )
        self.textLabel2.setText(_translate("ControlForm", "Port:"))
        self.linePort0.setToolTip(
            _translate("ControlForm", "Specify the port number of which to connect.")
        )
        self.groupBox_2.setTitle(_translate("ControlForm", "Backup Options"))
        self.rdbActive0.setToolTip(
            _translate(
                "ControlForm",
                (
                    "Only back up the active elements, buffered programs and coordinate"
                    " system axes definitions."
                ),
            )
        )
        self.rdbActive0.setText(_translate("ControlForm", "active"))
        self.rdbProject0.setToolTip(
            _translate(
                "ControlForm",
                "Only backup the saved and active Project files.",
            )
        )
        self.rdbProject0.setText(_translate("ControlForm", "project"))
        self.rdbAll0.setToolTip(_translate("ControlForm", "Back up everything."))
        self.rdbAll0.setText(_translate("ControlForm", "all"))
        self.label.setText(_translate("ControlForm", "Ignore file:"))
        self.lineIgnoreFile0.setToolTip(
            _translate(
                "ControlForm",
                "Path to the file listing which data structures should be ignored.",
            )
        )
        self.pushIgnoreLocate.setText(_translate("ControlForm", "Browse"))
        self.label_2.setText(_translate("ControlForm", "Output directory:"))
        self.lineOutputDir0.setToolTip(
            _translate(
                "ControlForm", "Directory in which the Backup and log will be written."
            )
        )
        self.pushDirLocate.setText(_translate("ControlForm", "Browse"))
        self.groupBox_3.setTitle(_translate("ControlForm", "Run"))
        self.pushBackup.setToolTip(
            _translate("ControlForm", "Backup the Power PMAC to a local file.")
        )
        self.pushBackup.setText(_translate("ControlForm", "Backup"))
        self.pushCancel0.setToolTip(
            _translate("ControlForm", "Cancel the current action.")
        )
        self.pushCancel0.setText(_translate("ControlForm", "Cancel"))
        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.Backup), _translate("ControlForm", "Backup")
        )
        self.groupBox_5.setTitle(
            _translate("ControlForm", "Power PMAC Compare Configuration")
        )
        self.groupBox_6.setTitle(_translate("ControlForm", "Compare sources"))
        self.textLabel1_2.setText(_translate("ControlForm", "Source 1:"))
        self.lineSource1.setToolTip(
            _translate(
                "ControlForm",
                (
                    "<html><head/><body><p>Source to compare can be:</p><p>  - path to"
                    "a back-up directory</p><p>  - a network interface"
                    " &lt;ip.address&gt;:&lt;port&gt; </p></body></html>"
                ),
            )
        )
        self.pushDirLocate_4.setText(_translate("ControlForm", "Browse"))
        self.textLabel1_3.setText(_translate("ControlForm", "Source 2:"))
        self.lineSource2.setToolTip(
            _translate(
                "ControlForm",
                (
                    "<html><head/><body><p>Source to compare can be:</p><p>  - path to"
                    " a back-up directory</p><p>  - a network interface"
                    " &lt;ip.address&gt;:&lt;port&gt; </p></body></html>"
                ),
            )
        )
        self.pushDirLocate_3.setText(_translate("ControlForm", "Browse"))
        self.groupBox_7.setTitle(_translate("ControlForm", "Compare Options"))
        self.label_3.setText(_translate("ControlForm", "Ignore file:"))
        self.lineIgnoreFile1.setToolTip(
            _translate(
                "ControlForm",
                "Path to the file listing which data structures should be ignored.",
            )
        )
        self.pushIgnoreLocate_2.setText(_translate("ControlForm", "Browse"))
        self.label_4.setText(_translate("ControlForm", "Output directory:"))
        self.lineOutputDir1.setToolTip(
            _translate(
                "ControlForm",
                "Directory in which the Compare information and log will be written.",
            )
        )
        self.pushDirLocate_2.setText(_translate("ControlForm", "Browse"))
        self.rdbActive1.setToolTip(
            _translate(
                "ControlForm",
                (
                    "Only back up the active elements, buffered programs and coordinate"
                    " system axes definitions."
                ),
            )
        )
        self.rdbActive1.setText(_translate("ControlForm", "active"))
        self.rdbProject1.setToolTip(
            _translate("ControlForm", "Only backup the saved and active Project files.")
        )
        self.rdbProject1.setText(_translate("ControlForm", "project"))
        self.rdbAll1.setToolTip(_translate("ControlForm", "Back up everything."))
        self.rdbAll1.setText(_translate("ControlForm", "all"))
        self.pushOpenViewer.setToolTip(
            _translate(
                "ControlForm",
                (
                    "After running a compare, open a window to display differences"
                    " between Source1 and Source2"
                ),
            )
        )
        self.pushOpenViewer.setText(_translate("ControlForm", "Open Comparison Viewer"))
        self.groupBox_8.setTitle(_translate("ControlForm", "Run"))
        self.pushCompare.setToolTip(
            _translate("ControlForm", "Compare two sources and save the output.")
        )
        self.pushCompare.setText(_translate("ControlForm", "Compare"))
        self.pushCancel1.setToolTip(
            _translate("ControlForm", "Cancel the current action.")
        )
        self.pushCancel1.setText(_translate("ControlForm", "Cancel"))
        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.tab_2), _translate("ControlForm", "Compare")
        )
        self.groupBox_9.setTitle(
            _translate("ControlForm", "Power PMAC Download/Recover Configuration")
        )
        self.groupBox_10.setTitle(_translate("ControlForm", "Connection to Power PMAC"))
        self.textLabel1_4.setText(_translate("ControlForm", "Server:"))
        self.lineServer2.setToolTip(
            _translate(
                "ControlForm",
                "Specify the IP address of the Power PMAC to download to.",
            )
        )
        self.textLabel2_2.setText(_translate("ControlForm", "Port:"))
        self.linePort2.setToolTip(
            _translate("ControlForm", "Specify the port number of which to connect.")
        )
        self.groupBox_11.setTitle(_translate("ControlForm", "Download/Recover Options"))
        self.label_5.setText(_translate("ControlForm", "Backup location:"))
        self.lineBackupDir.setToolTip(
            _translate(
                "ControlForm",
                "Location of the backup (output directory from running a Backup).",
            )
        )
        self.pushBackupDir.setText(_translate("ControlForm", "Browse"))
        self.label_6.setText(_translate("ControlForm", "Output directory:"))
        self.lineOutputDir2.setToolTip(
            _translate("ControlForm", "Output directory for the log file.")
        )
        self.pushOutputDir.setText(_translate("ControlForm", "Browse"))
        self.groupBox_12.setTitle(_translate("ControlForm", "Run"))
        self.pushDownload.setToolTip(
            _translate(
                "ControlForm",
                (
                    "<html><head/><body><p>Copied the specified built projects into the"
                    " active project site of the Power PMAC "
                    "(/var/ftp/usrflash/Project) and then loads it into active"
                    " memory.</p></body></html>"
                ),
            )
        )
        self.pushDownload.setText(_translate("ControlForm", "Download"))
        self.pushCancel2.setToolTip(
            _translate("ControlForm", "Cancel the current action.")
        )
        self.pushCancel2.setText(_translate("ControlForm", "Cancel"))
        self.pushRecover.setToolTip(
            _translate(
                "ControlForm",
                (
                    "<html><head/><body><p>Restore a previous saved configuration."
                    " Project, Database and Temp directories which have been"
                    " copied from the /opt/ppmac/usrflash site on the Power PMAC"
                    " are copied back into that site and the Project will become"
                    " active upon power cycling the"
                    " Power PMAC.</p></body></html>"
                ),
            )
        )
        self.pushRecover.setText(_translate("ControlForm", "Recover"))
        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.tab),
            _translate("ControlForm", "Download/Recover"),
        )
